from unittest import TestCase
from arcana.exceptions import (
    ArcanaError, ArcanaVersionError)
from arcana.environment.requirement import (
    PythonPackageRequirement, Version)
from arcana.environment.requirement.matlab import matlab_req, MatlabVersion


a_req = PythonPackageRequirement('a')
a_req_rnge = a_req.v('2.0.1', '10.2.1')
b_req = PythonPackageRequirement('b')
b_req_rnge = b_req.v('0.0.9', '1.0.1')
c_req = PythonPackageRequirement('c')
c_req_rnge = c_req.v('1.8.13', '2.22.0')


class TestRequirements(TestCase):

    def test_version_split(self):
        self.assertEqual(a_req.v('12.2.9').sequence, (12, 2, 9))
        a_ver = a_req.v('0.1a2')
        self.assertEqual(a_ver.sequence, (0, 1))
        self.assertEqual(a_ver.prerelease, ('a', 2))
        a_ver = a_req.v('0.1.3a2')
        self.assertEqual(a_ver.sequence, (0, 1, 3))
        self.assertEqual(a_ver.prerelease, ('a', 2))
        a_ver = a_req.v('0.1.3beta4')
        self.assertEqual(a_ver.sequence, (0, 1, 3))
        self.assertEqual(a_ver.prerelease, ('b', 4))
        a_ver = a_req.v('0.1.3beta4.dev12')
        self.assertEqual(a_ver.sequence, (0, 1, 3))
        self.assertEqual(a_ver.prerelease, ('b', 4))
        self.assertEqual(a_ver.dev, 12)
        a_ver = a_req.v('4.0.5_RC10')
        self.assertEqual(a_ver.sequence, (4, 0, 5))
        self.assertEqual(a_ver.prerelease, ('rc', 10))
        a_ver = a_req.v('4.0.5rc2')
        self.assertEqual(a_ver.sequence, (4, 0, 5))
        self.assertEqual(a_ver.prerelease, ('rc', 2))
        self.assertEqual(matlab_req.v('2015b').sequence,
                         (2015, 'b'))
        self.assertEqual(matlab_req.v('r2014a').sequence,
                         (2014, 'a'))
        self.assertEqual(matlab_req.v('R2017b').sequence,
                         (2017, 'b'))
        self.assertEqual(matlab_req.v('R2017B').sequence, (2017, 'b'))

    def test_latest_version(self):
        self.assertEqual(
            a_req_rnge.latest_within([a_req.v('2.0.0'), a_req.v('2.0.1')]),
            a_req.v('2.0.1'))
        self.assertEqual(
            a_req_rnge.latest_within([a_req.v('2.0.0'), a_req.v('3.0.9')]),
            a_req.v('3.0.9'))
        self.assertEqual(
            b_req_rnge.latest_within([b_req.v('0.0.8'), b_req.v('0.0.9')]),
            b_req.v('0.0.9'))
        self.assertEqual(
            c_req_rnge.latest_within([c_req.v('1.9.4'), c_req.v('1.2.9'),
                                      c_req.v('3.0.1')]),
            c_req.v('1.9.4'))

    def test_exceptions(self):
        req = PythonPackageRequirement('anything')
        self.assertRaises(
            ArcanaError,
            req.v,
            '2.1.10',
            '2.0.11')
        self.assertRaises(
            ArcanaVersionError,
            a_req_rnge.latest_within,
            ['2.0.0', '1.9.1'])
        self.assertRaises(
            ArcanaVersionError,
            b_req_rnge.latest_within,
            ['2.2.0', '3.4.1'])



import re
from .base import BaseRequirement, Version
from arcana.utils import run_matlab_cmd
from arcana.exceptions import (
    ArcanaVersionNotDetectableError, ArcanaRequirementNotFoundError)


class MatlabVersion(Version):
    """
    A representation of the Matlab release, e.g. R2017b
    """

    @classmethod
    def parse(cls, version_str):
        regex = r'R?(\d+)(a|b)'
        match = re.search(regex, version_str, re.IGNORECASE)
        if match is None:
            raise ArcanaVersionNotDetectableError(
                "Could not parse Matlab version string {} as {}. Regex ({})"
                " did not match any sub-string".format(
                    version_str, cls.__name__, regex))
        return (int(match.group(1)), match.group(2).lower()), None, None, None

    def __str__(self):
        return 'R{}{}'.format(*self.sequence)


class MatlabRequirement(BaseRequirement):

    def detect_version_str(self):
        """
        Finds the version of Matlab requirement that is accessible in
        the current environment.
        """
        # The matlab version should be included in the opening splash
        return run_matlab_cmd("version('-release')")


matlab_req = MatlabRequirement(
    'matlab', version_cls=MatlabVersion)


class MatlabPackageRequirement(BaseRequirement):
    """
    Defines a software package within Matlab

    Parameters
    ----------
    name : str
        Name of the package
    min_version : tuple(int|str)
        The minimum version required by the node
    max_version : tuple(int|str) | None
        The maximum version that is compatible with the Node
    test_func : str
        The name of a function that should be available when the
        requirement is satisfied.
    version_split : function
        A function that splits the version string into major/minor/micro
        parts or equivalent
    """

    def __init__(self, name, test_func, **kwargs):
        super(MatlabPackageRequirement, self).__init__(name, **kwargs)
        self._test_func = test_func

    def __eq__(self, other):
        return (super(MatlabPackageRequirement, self).__eq__(other) and
                self._test_func == other._test_func)

    def __hash__(self):
        return super().__hash__() ^ hash(self._test_func)

    @property
    def test_func(self):
        return self._test_func

    def detect_version_str(self):
        """
        Try to detect version of package from command help text. Bit of a long
        shot as they are typically included
        """
        help_text = run_matlab_cmd("help('{}')".format(self.test_func))
        if not help_text:
            raise ArcanaRequirementNotFoundError(
                "Did not find test function '{}' for {}"
                .format(self.test_func, self))
        return self.parse_help_text(help_text)

    def parse_help_text(self, help_text):
        return help_text



from builtins import object
from collections import defaultdict
from abc import ABCMeta, abstractmethod
from future.utils import with_metaclass
import logging
from .tree import Tree
from arcana.exceptions import ArcanaUsageError


logger = logging.getLogger('arcana')


class Repository(with_metaclass(ABCMeta, object)):
    """
    Abstract base class for all Repository systems, DaRIS, XNAT and
    local file system. Sets out the interface that all Repository
    classes should implement.
    """

    def __init__(self, subject_id_map=None, visit_id_map=None,
                 file_formats=()):
        self._connection_depth = 0
        self._subject_id_map = subject_id_map
        self._visit_id_map = visit_id_map
        self._inv_subject_id_map = {}
        self._inv_visit_id_map = {}
        self._file_formats = file_formats
        self.clear_cache()

    def __enter__(self):
        # This allows the repository to be used within nested contexts
        # but still only use one connection. This is useful for calling
        # methods that need connections, and therefore control their
        # own connection, in batches using the same connection by
        # placing the batch calls within an outer context.
        if self._connection_depth == 0:
            self.connect()
        self._connection_depth += 1
        return self

    def __exit__(self, exception_type, exception_value, traceback):  # @UnusedVariable @IgnorePep8
        self._connection_depth -= 1
        if self._connection_depth == 0:
            self.disconnect()

    def connect(self):
        """
        If a connection session is required to the repository,
        manage it here
        """

    def disconnect(self):
        """
        If a connection session is required to the repository,
        manage it here
        """

    @abstractmethod
    def find_data(self, subject_ids=None, visit_ids=None, **kwargs):
        """
        Find all data within a repository, registering filesets, fields and
        provenance with the found_fileset, found_field and found_provenance
        methods, respectively

        Parameters
        ----------
        subject_ids : list(str)
            List of subject IDs with which to filter the tree with. If
            None all are returned
        visit_ids : list(str)
            List of visit IDs with which to filter the tree with. If
            None all are returned

        Returns
        -------
        filesets : list[Fileset]
            All the filesets found in the repository
        fields : list[Field]
            All the fields found in the repository
        records : list[Record]
            The provenance records found in the repository
        """

    @abstractmethod
    def get_fileset(self, fileset):
        """
        Cache the fileset locally if required

        Parameters
        ----------
        fileset : Fileset
            The fileset to cache locally

        Returns
        -------
        path : str
            The file-system path to the cached file
        """

    @abstractmethod
    def get_field(self, field):
        """
        Extract the value of the field from the repository

        Parameters
        ----------
        field : Field
            The field to retrieve the value for

        Returns
        -------
        value : int | float | str | list[int] | list[float] | list[str]
            The value of the Field
        """

    def get_checksums(self, fileset):  # @UnusedVariable
        """
        Returns the checksums for the files in the fileset that are stored in
        the repository. If no checksums are stored in the repository then this
        method should be left to return None and the checksums will be
        calculated by downloading the files and taking calculating the digests

        Parameters
        ----------
        fileset : Fileset
            The fileset to return the checksums for

        Returns
        -------
        checksums : dct[str, str]
            A dictionary with keys corresponding to the relative paths of all
            files in the fileset from the base path and values equal to the MD5
            hex digest. The primary file in the file-set (i.e. the one that the
            path points to) should be specified by '.'.
        """
        return None

    @abstractmethod
    def put_fileset(self, fileset):
        """
        Inserts or updates the fileset into the repository

        Parameters
        ----------
        fileset : Fileset
            The fileset to insert into the repository
        """

    @abstractmethod
    def put_field(self, field):
        """
        Inserts or updates the fields into the repository

        Parameters
        ----------
        field : Field
            The field to insert into the repository
        """

    @abstractmethod
    def put_record(self, record):
        """
        Inserts a provenance record into a session or subject|visit|study
        summary
        """

    def tree(self, subject_ids=None, visit_ids=None, **kwargs):
        """
        Return the tree of subject and sessions information within a
        project in the XNAT repository

        Parameters
        ----------
        subject_ids : list(str)
            List of subject IDs with which to filter the tree with. If
            None all are returned
        visit_ids : list(str)
            List of visit IDs with which to filter the tree with. If
            None all are returned

        Returns
        -------
        tree : arcana.repository.Tree
            A hierarchical tree of subject, session and fileset
            information for the repository
        """
        # Find all data present in the repository (filtered by the passed IDs)
        return Tree.construct(
            self, *self.find_data(subject_ids=subject_ids,
                                  visit_ids=visit_ids), **kwargs)

    def cached_tree(self, subject_ids=None, visit_ids=None, fill=False):
        """
        Access the repository tree and caches it for subsequent
        accesses

        Parameters
        ----------
        subject_ids : list(str)
            List of subject IDs with which to filter the tree with. If
            None all are returned
        visit_ids : list(str)
            List of visit IDs with which to filter the tree with. If
            None all are returned
        fill : bool
            Create empty sessions for any that are missing in the
            subject_id x visit_id block. Typically only used if all
            the inputs to the study are coming from different repositories
            to the one that the derived products are stored in

        Returns
        -------
        tree : arcana.repository.Tree
            A hierarchical tree of subject, vist, session information and that
            of the filesets and fields they contain
        """
        if subject_ids is not None:
            subject_ids = frozenset(subject_ids)
        if visit_ids is not None:
            visit_ids = frozenset(visit_ids)
        try:
            tree = self._cache[subject_ids][visit_ids]
        except KeyError:
            if fill:
                fill_subjects = subject_ids
                fill_visits = visit_ids
            else:
                fill_subjects = fill_visits = None
            tree = self.tree(
                subject_ids=subject_ids, visit_ids=visit_ids,
                fill_visits=fill_visits, fill_subjects=fill_subjects)
            # Save the tree within the cache under the given subject/
            # visit ID filters and the IDs that were actually returned
            self._cache[subject_ids][visit_ids] = self._cache[
                frozenset(tree.subject_ids)][frozenset(tree.visit_ids)] = tree
        return tree

    def clear_cache(self):
        self._cache = defaultdict(dict)

    def __ne__(self, other):
        return not (self == other)

    def map_subject_id(self, subject_id):
        return self._map_id(subject_id, self._subject_id_map,
                            self._inv_subject_id_map)

    def map_visit_id(self, visit_id):
        return self._map_id(visit_id, self._visit_id_map,
                            self._inv_visit_id_map)

    def inv_map_subject_id(self, subject_id):
        try:
            return self._inv_subject_id_map[subject_id]
        except KeyError:
            return subject_id

    def inv_map_visit_id(self, visit_id):
        try:
            return self._inv_visit_id_map[visit_id]
        except KeyError:
            return visit_id

    def _map_id(self, id, map, inv_map):  # @ReservedAssignment
        if id is None:
            return None
        mapped = id
        if callable(map):
            mapped = map(id)
        elif self._visit_id_map is not None:
            try:
                mapped = map(id)
            except KeyError:
                pass
        if mapped != id:
            # Check for multiple mappings onto the same ID
            try:
                prev = inv_map[mapped]
            except KeyError:
                inv_map[mapped] = id
            else:
                if prev != id:
                    raise ArcanaUsageError(
                        "Both '{}' and '{}' have been mapped onto the same ID "
                        "in repository {}"
                        .format(prev, id, self))
        return mapped



from builtins import zip
import math
import os
from arcana.exceptions import (
    ArcanaError, ArcanaJobSubmittedException)
from .base import Processor
from nipype.pipeline.plugins.slurmgraph import SLURMGraphPlugin


class ArcanaSlurmGraphPlugin(SLURMGraphPlugin):

    def __init__(self, *args, **kwargs):
        self._processor = kwargs.pop('processor')
        super(ArcanaSlurmGraphPlugin, self).__init__(*args, **kwargs)

    def _get_args(self, node, keywords):
        """
        Intercept calls to get template and return our own node-specific
        template
        """
        args = super(ArcanaSlurmGraphPlugin, self)._get_args(
            node, keywords)
        # Substitute the template arg with the node-specific one
        new_args = []
        for name, arg in zip(keywords, args):
            if name == 'template':
                new_args.append(self._processor.slurm_template(node))
            else:
                new_args.append(arg)
        return tuple(new_args)


class SlurmProc(Processor):
    """
    A thin wrapper around the NiPype SLURMGraphPlugin used to connect
    submit pipelines to a Slurm scheduler

    Parameters
    ----------
    work_dir : str
        A directory in which to run the nipype workflows
    partition : str | function
        Either a string naming the partition to use for the jobs or a function
        that takes a node as an argument and returns the name of the partition
        to use for that particular node.
    account : str
        Name of the account to submit the jobs against.
    email : str | None
        The email address to send alerts to. If not provided the 'EMAIL'
        environment variable needs to be set
    generic_resources : function
        A function that takes a node as an argument and returns the list of
        generic resources (e.g. GPU, bandwidth) required
    mail_on : List[str]
        Conditions on which to send mail (default 'FAIL')
    max_process_time : float
        The maximum time allowed for the process
    reprocess: True|False|'all'
        A flag which determines whether to rerun the processing for this
        step. If set to 'all' then pre-requisite pipelines will also be
        reprocessed.
    """

    nipype_plugin_cls = ArcanaSlurmGraphPlugin

    def __init__(self, work_dir, partition=None, account=None, email=None,
                 mail_on=('FAIL',), generic_resources=None,
                 ntasks_per_node=None, cpus_per_task=None, **kwargs):
        if email is None:
            try:
                email = os.environ['EMAIL']
            except KeyError:
                raise ArcanaError(
                    "'email' kwarg needs to be provided for SlurmProc"
                    " if 'EMAIL' environment variable not set")
        self._email = email
        self._mail_on = mail_on
        self._account = account
        self._partition = partition
        self._ntasks_per_node = ntasks_per_node
        self._cpus_per_task = cpus_per_task
        self._generic_resources = generic_resources
        super(SlurmProc, self).__init__(work_dir, **kwargs)

    def _init_plugin(self):
        self._plugin = self.nipype_plugin_cls(processor=self,
                                              **self._plugin_args)

    @property
    def email(self):
        return self._email

    @property
    def mail_on(self):
        return self._mail_on

    @property
    def account(self):
        return self._account

    def run(self, pipeline, **kwargs):
        super(SlurmProc, self).run(pipeline, **kwargs)
        raise ArcanaJobSubmittedException(
            "Pipeline '{}' has been submitted to SLURM scheduler "
            "for processing. Please run script again after the jobs "
            "have been successful.")

    def slurm_template(self, node):
        sbatch = self.sbatch_template.format(
            wall_time=self.wall_time_str(node.wall_time), ntasks=node.n_procs,
            memory=int(node.mem_gb * 1000),
            email=self.email,
            account=self.account)
        if self.account is not None:
            sbatch += ("\n# Set the account\n"
                       "#SBATCH --account={}\n".format(self.account))
        if self._partition is not None:
            sbatch += ("\n# Set the partition to run the job on\n"
                       "#SBATCH --partition={}\n".format(
                           self._partition(node) if callable(self._partition)
                           else self._partition))
        if self._generic_resources is not None:
            sbatch += ("\n# Request generic resources\n")
            for gres in self._generic_resources(node):
                sbatch += '#SBATCH --gres={}\n'.format(gres)
        if self._mail_on:
            sbatch += ("\n# Set mail triggers\n")
            for mo in self._mail_on:
                sbatch += '#SBATCH --mail-type={}\n'.format(mo)
        sbatch += "\n# Node and CPU options\n"
        if self._cpus_per_task is not None:
            sbatch += "#SBATCH --cpus-per-task={}\n".format(
                self._cpus_per_task)
        if self._ntasks_per_node is not None:
            sbatch += "#SBATCH --ntasks-per-node={}\n".format(
                self._ntasks_per_node)
        return sbatch

    def wall_time_str(self, wall_time):
        """
        Returns the wall time in the format required for the sbatch script
        """
        days = int(wall_time // 1440)
        hours = int((wall_time - days * 1440) // 60)
        minutes = int(math.floor(wall_time - days * 1440 - hours * 60))
        seconds = int((wall_time - math.floor(wall_time)) * 60)
        return "{}-{:0>2}:{:0>2}:{:0>2}".format(days, hours, minutes, seconds)

    sbatch_template = """#!/bin/bash

# Set the email
#SBATCH --email={email}

# Request CPU resource for a parallel job
#SBATCH --ntasks={ntasks}

# Memory usage (MB)
#SBATCH --mem-per-cpu={memory}

# Set your minimum acceptable walltime, format: day-hours:minutes:seconds
#SBATCH --time={wall_time}

# Kill job if dependencies fail
#SBATCH --kill-on-invalid-dep=yes
"""



from __future__ import absolute_import
import os.path as op
import tempfile
import unittest
import sys
from unittest import TestCase
import xnat
from arcana.utils.testing import BaseTestCase
from arcana.data import InputFileset
from arcana.data.file_format import text_format
from arcana.repository import XnatRepo
from arcana.processor import SingleProc
from arcana.utils.testing.xnat import SKIP_ARGS, SERVER, TestOnXnatMixin


# Import TestExistingPrereqs study to test it on XNAT
sys.path.insert(0, op.join(op.dirname(__file__), '..', '..'))
import test_data  # @UnresolvedImport @IgnorePep8
from test_data import dicom_format  # @UnresolvedImport @IgnorePep8
sys.path.pop(0)

# Import test_local to run TestProjectInfo on XNAT using TestOnXnat mixin
sys.path.insert(0, op.join(op.dirname(__file__), '..', '..', 'processor'))
import test_to_process  # @UnresolvedImport @IgnorePep8
sys.path.pop(0)


class TestConnectDisconnect(TestCase):

    @unittest.skipIf(*SKIP_ARGS)
    def test_connect_disconnect(self):
        repository = XnatRepo(project_id='dummy',
                              server=SERVER,
                              cache_dir=tempfile.mkdtemp())
        with repository:
            self._test_open(repository)
        self._test_closed(repository)

        with repository:
            self._test_open(repository)
            with repository:
                self._test_open(repository)
            self._test_open(repository)
        self._test_closed(repository)

    def _test_open(self, repository):
        repository._login.classes  # check connection

    def _test_closed(self, repository):
        self.assertRaises(
            AttributeError,
            getattr,
            repository._login,
            'classes')


class TestProvInputChangeOnXnat(TestOnXnatMixin,
                                test_to_process.TestProvInputChange):

    BASE_CLASS = test_to_process.TestProvInputChange

    @unittest.skipIf(*SKIP_ARGS)
    def test_input_change(self):
        super(TestProvInputChangeOnXnat, self).test_input_change()


class TestDicomTagMatchAndIDOnXnat(TestOnXnatMixin,
                                   test_data.TestDicomTagMatch):

    BASE_CLASS = test_data.TestDicomTagMatch
    REF_FORMATS = [dicom_format]

    @property
    def ref_dir(self):
        return op.join(
            self.ref_path, self._get_name(self.BASE_CLASS))

    def setUp(self):
        test_data.TestDicomTagMatch.setUp(self)
        TestOnXnatMixin.setUp(self)
        # Set up DICOM headers
        with xnat.connect(SERVER) as login:
            xsess = login.projects[self.project].experiments[
                '_'.join((self.project, self.SUBJECT, self.VISIT))]
            login.put('/data/experiments/{}?pullDataFromHeaders=true'
                      .format(xsess.id))

    def tearDown(self):
        TestOnXnatMixin.tearDown(self)
        test_data.TestDicomTagMatch.tearDown(self)

    @unittest.skipIf(*SKIP_ARGS)
    def test_dicom_match(self):
        study = test_data.TestMatchStudy(
            name='test_dicom',
            repository=XnatRepo(
                project_id=self.project,
                server=SERVER, cache_dir=tempfile.mkdtemp()),
            processor=SingleProc(self.work_dir),
            inputs=test_data.TestDicomTagMatch.DICOM_MATCH)
        phase = list(study.data('gre_phase'))[0]
        mag = list(study.data('gre_mag'))[0]
        self.assertEqual(phase.name, 'gre_field_mapping_3mm_phase')
        self.assertEqual(mag.name, 'gre_field_mapping_3mm_mag')

    @unittest.skipIf(*SKIP_ARGS)
    def test_id_match(self):
        study = test_data.TestMatchStudy(
            name='test_dicom',
            repository=XnatRepo(
                project_id=self.project,
                server=SERVER, cache_dir=tempfile.mkdtemp()),
            processor=SingleProc(self.work_dir),
            inputs=[
                InputFileset('gre_phase', format=dicom_format, id=7),
                InputFileset('gre_mag', format=dicom_format, id=6)])
        phase = list(study.data('gre_phase'))[0]
        mag = list(study.data('gre_mag'))[0]
        self.assertEqual(phase.name, 'gre_field_mapping_3mm_phase')
        self.assertEqual(mag.name, 'gre_field_mapping_3mm_mag')

    @unittest.skipIf(*SKIP_ARGS)
    def test_order_match(self):
        test_data.TestDicomTagMatch.test_order_match(self)


class TestFilesetCacheOnPathAccess(TestOnXnatMixin,
                                   BaseTestCase):

    INPUT_FILESETS = {'fileset': '1'}

    @unittest.skipIf(*SKIP_ARGS)
    def test_cache_on_path_access(self):
        tmp_dir = tempfile.mkdtemp()
        repository = XnatRepo(
            project_id=self.project,
            server=SERVER, cache_dir=tmp_dir)
        tree = repository.tree(
            subject_ids=[self.SUBJECT],
            visit_ids=[self.VISIT])
        # Get a fileset
        fileset = list(list(list(tree.subjects)[0].sessions)[0].filesets)[0]
        fileset.format = text_format
        self.assertEqual(fileset._path, None)
        target_path = op.join(
            tmp_dir, self.project,
            '{}_{}'.format(self.project, self.SUBJECT),
            '{}_{}_{}'.format(self.project, self.SUBJECT, self.VISIT),
            fileset.basename, fileset.fname)
        # This should implicitly download the fileset
        self.assertEqual(fileset.path, target_path)
        with open(target_path) as f:
            self.assertEqual(f.read(),
                             self.INPUT_FILESETS[fileset.basename])



from __future__ import absolute_import
from future.utils import with_metaclass
import os.path as op
import tempfile
import unittest
from arcana.utils.testing import BaseMultiSubjectTestCase
from arcana.repository.xnat import XnatRepo
from arcana.data import (
    InputFileset, InputFilesetSpec)
from arcana.study import Study, StudyMetaClass
from arcana.data.file_format import text_format
from arcana.repository.tree import Tree, Subject, Session, Visit
from arcana.data import Fileset
import sys
from arcana.utils.testing.xnat import (
    TestMultiSubjectOnXnatMixin, SKIP_ARGS, SERVER)

# Import TestExistingPrereqs study to test it on XNAT
sys.path.insert(0, op.join(op.dirname(__file__), '..', '..', 'study'))
import test_study  # @UnresolvedImport @IgnorePep8
sys.path.pop(0)

# Import test_local to run TestProjectInfo on XNAT using TestOnXnat mixin
sys.path.insert(0, op.join(op.dirname(__file__), '..'))
import test_directory  # @UnresolvedImport @IgnorePep8
sys.path.pop(0)


class TestStudy(with_metaclass(StudyMetaClass, Study)):

    add_data_specs = [
        InputFilesetSpec('fileset1', text_format),
        InputFilesetSpec('fileset2', text_format, optional=True),
        InputFilesetSpec('fileset3', text_format),
        InputFilesetSpec('fileset5', text_format, optional=True)]


class TestExistingPrereqsOnXnat(TestMultiSubjectOnXnatMixin,
                                test_study.TestExistingPrereqs):

    BASE_CLASS = test_study.TestExistingPrereqs

    @unittest.skipIf(*SKIP_ARGS)
    def test_per_session_prereqs(self):
        super(TestExistingPrereqsOnXnat, self).test_per_session_prereqs()


class TestProjectInfo(TestMultiSubjectOnXnatMixin,
                      test_directory.TestDirectoryProjectInfo):

    BASE_CLASS = test_directory.TestDirectoryProjectInfo

    @unittest.skipIf(*SKIP_ARGS)
    def test_project_info(self):
        tree = self.repository.tree()
        for node in tree.nodes():
            for fileset in node.filesets:
                fileset.format = text_format
                # Clear id and format name from regenerated tree
                fileset._id = None
#                 fileset.get()
        ref_tree = self.get_tree(self.repository)  #, sync_with_repo=True)
        self.assertEqual(
            tree, ref_tree,
            "Generated project doesn't match reference:{}"
            .format(tree.find_mismatch(ref_tree)))


class TestXnatCache(TestMultiSubjectOnXnatMixin,
                    BaseMultiSubjectTestCase):

    BASE_CLASS = BaseMultiSubjectTestCase
    STRUCTURE = {
        'subject1': {
            'visit1': ['fileset1', 'fileset2', 'fileset3'],
            'visit2': ['fileset1', 'fileset2', 'fileset3']},
        'subject2': {
            'visit1': ['fileset1', 'fileset2', 'fileset3'],
            'visit2': ['fileset1', 'fileset2', 'fileset3']}}

    DATASET_CONTENTS = {'fileset1': 1,
                        'fileset2': 2,
                        'fileset3': 3}

    @property
    def input_tree(self):
        filesets = []
        for subj_id, visits in list(self.STRUCTURE.items()):
            for visit_id, fileset_names in list(visits.items()):
                filesets.extend(
                    Fileset(d, text_format, subject_id=subj_id,
                            visit_id=visit_id) for d in fileset_names)
        return Tree.construct(self.repository, filesets=filesets)

    @unittest.skipIf(*SKIP_ARGS)
    def test_cache_download(self):
        repository = XnatRepo(
            project_id=self.project,
            server=SERVER,
            cache_dir=tempfile.mkdtemp())
        study = self.create_study(
            TestStudy, 'cache_download',
            inputs=[
                InputFileset('fileset1', 'fileset1', text_format),
                InputFileset('fileset3', 'fileset3', text_format)],
            repository=repository)
        study.cache_inputs()
        for subject_id, visits in list(self.STRUCTURE.items()):
            subj_dir = op.join(
                repository.cache_dir, self.project,
                '{}_{}'.format(self.project, subject_id))
            for visit_id in visits:
                sess_dir = op.join(
                    subj_dir,
                    '{}_{}_{}'.format(self.project, subject_id,
                                      visit_id))
                for inpt in study.inputs:
                    self.assertTrue(op.exists(op.join(sess_dir, inpt.name)))

    @property
    def base_name(self):
        return self.name



# -*- coding: utf-8 -*-
from __future__ import division
from future.utils import PY3
from builtins import str  # @UnusedImport
import os
import re
import logging
import subprocess as sp
from collections import defaultdict
from arcana.exceptions import (
    ArcanaError, ArcanaModulesNotInstalledException,
    ArcanaRequirementNotFoundError, ArcanaVersionNotDetectableError,
    ArcanaVersionError)
from .base import Environment, NodeMixin, Node, JoinNode, MapNode


logger = logging.getLogger('arcana')


class ModulesNodeMixin(NodeMixin):

    def _run_command(self, *args, **kwargs):
        try:
            self.environment.load(*self.versions)
            result = self.base_cls._run_command(self, *args, **kwargs)
        finally:
            self.environment.unload(*self.versions)
        return result

    def _load_results(self, *args, **kwargs):
        self.environment.load(*self.versions)
        result = self.base_cls._load_results(self, *args, **kwargs)
        self.environment.unload(*self.versions)
        return result


class ModulesNode(ModulesNodeMixin, Node):

    base_cls = Node  # Not req. in Py3 where super() in mixin works


class ModulesJoinNode(ModulesNodeMixin, JoinNode):

    base_cls = JoinNode  # Not req. in Py3 where super() in mixin works


class ModulesMapNode(ModulesNodeMixin, MapNode):

    node_cls = ModulesNode
    base_cls = MapNode  # Not req. in Py3 where super() in mixin works


class ModulesEnv(Environment):
    """
    An environment in which software requirements (e.g. FSL, matlab,
    MRtrix) are loaded using the 'modules' package

        Furlani, J. (1991). Modules: Providing a flexible user
        environment. Proceedings of the Fifth Large Installation Systems
        Administration Conference (LISA V), (1), 141–152.

    Parameters
    ----------
    packages_map : dct[Requirement, str] | callable
        A Mapping from the name of a requirement to a module installed in the
        environment
    versions_map : dct[Requirement, dct[str, str]] | callable
        A dictionary containing a mapping from a Requirement to dictionary,
        which in turn maps the local name of a version of a module installed in
        the environment to the standard versioning system recognised by the
        requirement.
    ignore_unrecognised : bool
        If True, then unrecognisable versions are ignored instead of
        throwing an error
    """

    node_types = {'base': ModulesNode, 'map': ModulesMapNode,
                  'join': ModulesJoinNode}

    def __init__(self, packages_map=None, versions_map=None,
                 fail_on_missing=True, ignore_unrecognised=True,
                 detect_exact_versions=True):
        if packages_map is None:
            packages_map = {}
        if versions_map is None:
            versions_map = {}
        self._packages_map = packages_map
        self._versions_map = versions_map
        self._ignore_unrecog = ignore_unrecognised
        self._fail_on_missing = fail_on_missing
        self._detect_exact_versions = detect_exact_versions
        self._detected_cache = None
        self._available = self.available()

    def __eq__(self, other):
        return (self._packages_map == other._packages_map and
                self._versions_map == other._versions_map and
                self._fail_on_missing == other._fail_on_missing and
                self._ignore_unrecog == other._ignore_unrecog and
                self._detect_exact_version == other._detect_exact_version)

    def satisfy(self, *requirements):
        versions = []
        for req_range in requirements:
            req = req_range.requirement
            local_name = self.map_req(req)
            try:
                version_names = self._available[local_name]
            except KeyError:
                if self._fail_on_missing:
                    raise ArcanaRequirementNotFoundError(
                        "Could not find module for {} ({})".format(req.name,
                                                                   local_name))
                else:
                    logger.warning("Did not find module for {} ({})"
                                   .format(req.name, local_name))
            avail_versions = []
            for local_ver_name in version_names:
                ver_name = self.map_version(req, local_ver_name)
                try:
                    avail_versions.append(
                        req.v(ver_name, local_name=local_name,
                              local_version=local_ver_name))
                except ArcanaVersionNotDetectableError:
                    if self._ignore_unrecog:
                        logger.warning(
                            "Ignoring unrecognised available version '{}' of "
                            "{}".format(ver_name, req_range.name))
                        continue
                    else:
                        raise
            version = req_range.latest_within(avail_versions)
            # To get the exact version (i.e. not just what the
            # modules administrator has called it) we load the module
            # detect the version and unload it again
            if self._detect_exact_versions:
                self.load(version)
                exact_version = req.detect_version(
                    local_name=local_name, local_version=local_ver_name)
                self.unload(version)
                if not req_range.within(exact_version):
                    raise ArcanaVersionError(
                        "Version of {} specified by module {} does not match "
                        "expected {} and is outside the acceptable range [{}]"
                        .format(req.name, local_ver_name, str(version),
                                str(req_range)))
                if exact_version < version:
                    raise ArcanaVersionError(
                        "Version of {} specified by module {} is less than "
                        "the expected {}"
                        .format(req.name, local_ver_name, str(version)))
                version = exact_version
            # Get latest requirement from list of possible options
            versions.append(version)
        return versions

    def load(self, *versions):
        for version in versions:
            self._run_module_cmd('load', self._module_id(version))

    def unload(self, *versions):
        for version in versions:
            self._run_module_cmd('unload', self._module_id(version))

    @classmethod
    def _module_id(cls, version):
        return '{}/{}'.format(version.local_name, version.local_version)

    def map_req(self, requirement):
        """
        Maps the name of an Requirement class to the name of the corresponding
        module in the environment

        Parameters
        ----------
        requirement : Requirement
            The requirement to map to the name of a module on the system
        """
        if isinstance(self._packages_map, dict):
            local_name = self._packages_map.get(requirement, requirement.name)
        else:
            local_name = self._packages_map(requirement)
        return local_name

    def map_version(self, requirement, local_version):
        """
        Maps a local version name to one recognised by the Requirement class

        Parameters
        ----------
        requirement : str
            Name of the requirement
        version : str
            version string
        """
        if isinstance(self._versions_map, dict):
            version = self._versions_map.get(requirement, {}).get(
                local_version, local_version)
        else:
            version = self._versions_map(requirement, local_version)
        return version

    @classmethod
    def loaded(cls):
        loaded = {}
        loaded_str = os.environ.get('LOADEDMODULES', '')
        if loaded_str:
            for modstr in loaded_str.split(':'):
                parts = modstr.split('/')
                if len(parts) == 2:
                    name, versionstr = parts
                else:
                    name = parts[0]
                    versionstr = None
                loaded[name] = versionstr
        return loaded

    @classmethod
    def available(cls):
        out_text = cls._run_module_cmd('avail')
        sanitized = []
        for l in out_text.split('\n'):
            if not l.startswith('-'):
                sanitized.append(l)
        available = defaultdict(list)
        for module, ver in re.findall(r'(\w+)/([\w\d\.\-\_]+)',
                                      ' '.join(sanitized)):
            available[module.lower()].append(ver)
        return available

    @classmethod
    def _run_module_cmd(cls, *args):
        if 'MODULESHOME' in os.environ:
            try:
                modulecmd = sp.check_output('which modulecmd',
                                            shell=True).strip()
            except sp.CalledProcessError:
                modulecmd = False
            if not modulecmd:
                modulecmd = '{}/bin/modulecmd'.format(
                    os.environ['MODULESHOME'])
                if not os.path.exists(modulecmd):
                    raise ArcanaError(
                        "Cannot find 'modulecmd' on path or in MODULESHOME.")
            logger.debug("Running modules command '{}'".format(' '.join(args)))
            try:
                output, error = sp.Popen(
                    [modulecmd, 'python'] + list(args),
                    stdout=sp.PIPE, stderr=sp.PIPE).communicate()
            except (sp.CalledProcessError, OSError) as e:
                raise ArcanaError(
                    "Call to subprocess `{}` threw an error: {}".format(
                        ' '.join([modulecmd, 'python'] + list(args)), e))
            exec(output)
            if PY3:
                error = error.decode('utf-8')
            return error
        else:
            raise ArcanaModulesNotInstalledException('MODULESHOME')



from past.builtins import basestring
from future.utils import PY2
from builtins import object
import os
import sys
from copy import deepcopy, copy
import tempfile
import shutil
import errno
import json
from itertools import chain
from collections import defaultdict
import networkx.readwrite.json_graph as nx_json
from networkx import __version__ as networkx_version
from nipype.pipeline import engine as pe
from nipype.interfaces.utility import IdentityInterface
from logging import getLogger
from arcana.utils import extract_package_version
from arcana.__about__ import __version__
from arcana.exceptions import (
    ArcanaDesignError, ArcanaError, ArcanaUsageError, ArcanaNoConverterError)
from .provenance import (
    Record, ARCANA_DEPENDENCIES, PROVENANCE_VERSION)


logger = getLogger('arcana')


class Pipeline(object):
    """
    A fairly thin wrapper around a NiPype workflow, which keeps track
    inputs and outputs, and maps names to avoid name-clashes and allow
    specialisation

    Parameters
    ----------
    study : Study
        The study from which the pipeline was created
    name : str
        The name of the pipeline
    name_maps : dict
        A dictionary containing several modifying keyword arguments that
        manipulate way the pipeline is constructed (e.g. map inputs and outputs
        to new entries in the data specification table). Typically names of
        inputs, outputs and the pipeline itself. Intended to allow secondary
        pipeline constructors to call a constructor, and return a modified
        version of the pipeline it returns.

        It should be passed directly from wildcard keyword args passed to the
        pipeline constructor, e.g.

        def my_pipeline_constructor(**name_maps):
            pipeline = self.new_pipeline('my_pipeline', name_maps=name_maps)
            pipeline.add('a_node', MyInterface())

            ...

            return pipeline

        The keywords in 'name_maps' used in pipeline construction are:

        name : str
            A new name for the pipeline
        prefix : str
            Prefix prepended to the original name of the pipeline. Typically
            only one of name and prefix is used at each nested level, but they
            can be used in conjunction.
        input_map : str | dict[str,str]
            Applied to the input names used by the pipeline to map them to new
            entries of the data specification in modified pipeline
            constructors. Typically used in sub-class or multi-study. If a
            string, the map is interpreted as a prefix applied to the names
            given in the original pipeline, if it is a dictionary the names are
            mapped explicitly.
        output_map : str | dict[str,str]
            Same as the input map but applied to outputs instead of inputs to
            the pipeline.
        name_maps : dict
            Modifications from nested pipeline constructors
        study : Study
            A different study to bind the pipeline to from the one containing
            the inner pipeline constructor. Intended to be used with
            multi-studies.
    desc : str
        A description of what the pipeline does
    references : List[Citation]
        List of scientific papers that describe the workflow and should be
        cited in publications that use it
    """

    def __init__(self, study, name, name_maps, desc=None, citations=None):
        name, study, maps = self._unwrap_maps(name_maps, name, study=study)
        self._name = name
        self._input_map = maps.get('input_map', None)
        self._output_map = maps.get('output_map', None)
        self._study = study
        self._workflow = pe.Workflow(name=self.name)
        self._desc = desc
        self._input_conns = defaultdict(list)
        self._iterator_conns = defaultdict(list)
        self._output_conns = {}
        self._iterator_joins = set()
        # Set up inputs
        self._citations = citations if citations is not None else []
        # For recording which parameters are accessed
        # during pipeline generation so they can be attributed to the
        # pipeline after it is generated (and then saved in the
        # provenance
        self._required_outputs = set()
        # Create placeholders for expected provenance records that are used
        # to compare with records saved in the repository when checking for
        # mismatches
        self._prov = None
        self._inputnodes = None
        self._outputnodes = None

    def __repr__(self):
        return "{}(name='{}')".format(self.__class__.__name__,
                                      self.name)

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        return (
            self._name == other._name and
            self._desc == other._desc and
            set(self._input_conns.keys()) == set(other._input_conns.keys()) and
            (set(self._output_conns.keys()) ==
             set(other._output_conns.keys())) and
            self._references == other._references)

    def __hash__(self):
        return (hash(self._name) ^
                hash(self._desc) ^
                hash(tuple(self._input_conns.keys())) ^
                hash(tuple(self._output_conns.keys())) ^
                hash(tuple(self._references)))

    def __ne__(self, other):
        return not (self == other)

    @property
    def has_prerequisites(self):
        return any(self._study.spec(i).is_spec for i in self.inputs)

    @property
    def prerequisites(self):
        """
        Iterates through the inputs of the pipelinen and determines the
        all prerequisite pipelines
        """
        # Loop through the inputs to the pipeline and add the instancemethods
        # for the pipelines to generate each of the processed inputs
        prereqs = defaultdict(set)
        for input in self.inputs:  # @ReservedAssignment
            spec = self._study.spec(input)
            # Could be an input to the study or optional acquired spec
            if spec.is_spec and spec.derived:
                prereqs[spec.pipeline_getter].add(input.name)
        return prereqs

    @property
    def study_inputs(self):
        """
        Returns all inputs of the study used by the pipeline, including inputs
        of prerequisites (and their prerequisites recursively)
        """
        return set(chain(
            (i for i in self.inputs if not i.derived),
            *(self.study.pipeline(p, required_outputs=r).study_inputs
              for p, r in self.prerequisites.items())))

    def add(self, name, interface, inputs=None, outputs=None,
            requirements=None, wall_time=None, annotations=None, **kwargs):
        """
        Adds a processing Node to the pipeline

        Parameters
        ----------
        name : str
            Name for the node
        interface : nipype.Interface
            The interface to use for the node
        inputs : dict[str, (str, FileFormat) | (Node, str)]
            Connections from inputs of the pipeline and outputs of other nodes
            to inputs of node. The keys of the dictionary are the field names
            and the values are 2-tuple containing either the name of the data
            spec and the data format it is expected in for pipeline inputs or
            the sending Node and the the name of an output of the sending Node.
            Note that pipeline inputs can be specified outside this method
            using the 'connect_input' method and connections between nodes with
            the the 'connect' method.
        outputs : dict[str, (str, FileFormat)]
            Connections to outputs of the pipeline from fields of the
            interface. The keys of the dictionary are the names of the data
            specs that will be written to and the values are the interface
            field name and the data format it is produced in. Note that output
            connections can also be specified using the 'connect_output'
            method.
        requirements : list(Requirement)
            List of required packages need for the node to run (default: [])
        wall_time : float
            Time required to execute the node in minutes (default: 1)
        mem_gb : int
            Required memory for the node in GB
        n_procs : int
            Preferred number of threads to run the node on (default: 1)
        annotations : dict[str, *]
            Additional annotations to add to the node, which may be used by
            the Processor node to optimise execution (e.g. 'gpu': True)
        iterfield : str
            Name of field to be passed an iterable to iterator over.
            If present, a MapNode will be created instead of a regular node
        joinsource : str
            Name of iterator field to join. Typically one of the implicit
            iterators (i.e. Study.SUBJECT_ID or Study.VISIT_ID)
            to join over the subjects and/or visits
        joinfield : str
            Name of field to pass the joined list when creating a JoinNode

        Returns
        -------
        node : Node
            The Node object that has been added to the pipeline
        """
        if annotations is None:
            annotations = {}
        if requirements is None:
            requirements = []
        if wall_time is None:
            wall_time = self.study.processor.default_wall_time
        if 'mem_gb' not in kwargs or kwargs['mem_gb'] is None:
            kwargs['mem_gb'] = self.study.processor.default_mem_gb
        if 'iterfield' in kwargs:
            if 'joinfield' in kwargs or 'joinsource' in kwargs:
                raise ArcanaDesignError(
                    "Cannot provide both joinsource and iterfield to when "
                    "attempting to add '{}' node to {}"
                    .foramt(name, self._error_msg_loc))
            node_cls = self.study.environment.node_types['map']
        elif 'joinsource' in kwargs or 'joinfield' in kwargs:
            if not ('joinfield' in kwargs and 'joinsource' in kwargs):
                raise ArcanaDesignError(
                    "Both joinsource and joinfield kwargs are required to "
                    "create a JoinNode (see {})".format(name,
                                                        self._error_msg_loc))
            joinsource = kwargs['joinsource']
            if joinsource in self.study.ITERFIELDS:
                self._iterator_joins.add(joinsource)
            node_cls = self.study.environment.node_types['join']
            # Prepend name of pipeline of joinsource to match name of nodes
            kwargs['joinsource'] = '{}_{}'.format(self.name, joinsource)
        else:
            node_cls = self.study.environment.node_types['base']
        # Create node
        node = node_cls(self.study.environment,
                        interface,
                        name="{}_{}".format(self._name, name),
                        requirements=requirements,
                        wall_time=wall_time,
                        annotations=annotations,
                        **kwargs)
        # Ensure node is added to workflow
        self._workflow.add_nodes([node])
        # Connect inputs, outputs and internal connections
        if inputs is not None:
            assert isinstance(inputs, dict)
            for node_input, connect_from in inputs.items():
                if isinstance(connect_from[0], basestring):
                    input_spec, input_format = connect_from
                    self.connect_input(input_spec, node,
                                       node_input, input_format)
                else:
                    conn_node, conn_field = connect_from
                    self.connect(conn_node, conn_field, node, node_input)
        if outputs is not None:
            assert isinstance(outputs, dict)
            for output_spec, (node_output, output_format) in outputs.items():
                self.connect_output(output_spec, node, node_output,
                                    output_format)
        return node

    def connect_input(self, spec_name, node, node_input, format=None, **kwargs):  # @ReservedAssignment @IgnorePep8
        """
        Connects a study fileset_spec as an input to the provided node

        Parameters
        ----------
        spec_name : str
            Name of the study data spec (or one of the IDs from the iterator
            nodes, 'subject_id' or 'visit_id') to connect to the node
        node : arcana.Node
            The node to connect the input to
        node_input : str
            Name of the input on the node to connect the fileset spec to
        format : FileFormat | None
            The file format the input is expected in. If it differs
            from the format in data spec or of study input then an implicit
            conversion is performed. If None the file format in the data spec
            is assumed
        """
        if spec_name in self.study.ITERFIELDS:
            self._iterator_conns[spec_name].append((node, node_input, format))
        else:
            name = self._map_name(spec_name, self._input_map)
            if name not in self.study.data_spec_names():
                raise ArcanaDesignError(
                    "Proposed input '{}' to {} is not a valid spec name ('{}')"
                    .format(name, self._error_msg_loc,
                            "', '".join(self.study.data_spec_names())))
            self._input_conns[name].append((node, node_input, format, kwargs))

    def connect_output(self, spec_name, node, node_output, format=None,   # @ReservedAssignment @IgnorePep8
                       **kwargs):
        """
        Connects an output to a study fileset spec

        Parameters
        ----------
        spec_name : str
            Name of the study fileset spec to connect to
        node : arcana.Node
            The node to connect the output from
        node_output : str
            Name of the output on the node to connect to the fileset
        format : FileFormat | None
            The file format the output is returned in. If it differs
            from the format in data spec then an implicit conversion is
            performed. If None the it is assumed to be returned in the file
            format of the entry the data spec
        """
        name = self._map_name(spec_name, self._output_map)
        if name not in self.study.data_spec_names():
            raise ArcanaDesignError(
                "Proposed output '{}' to {} is not a valid spec name ('{}')"
                .format(name, self._error_msg_loc,
                        "', '".join(self.study.data_spec_names())))
        if name in self._output_conns:
            prev_node, prev_node_output, _, _ = self._output_conns[name]
            logger.info(
                "Reassigning '{}' output from {}:{} to {}:{} in {}"
                .format(name, prev_node.name, prev_node_output,
                        node.name, node_output, self._error_msg_loc))
        self._output_conns[name] = (node, node_output, format, kwargs)

    def _map_name(self, name, mapper):
        """
        Maps a spec name to a new value based on the provided mapper
        """
        if mapper is not None:
            if isinstance(mapper, basestring):
                name = mapper + name
            try:
                name = mapper[name]
            except KeyError:
                pass
        return name

    def connect(self, *args, **kwargs):
        """
        Performs the connection in the wrapped NiPype workflow
        """
        self._workflow.connect(*args, **kwargs)

    @property
    def name(self):
        return self._name

    @property
    def study(self):
        return self._study

    @property
    def workflow(self):
        return self._workflow

    @property
    def references(self):
        return self._references

    @property
    def inputs(self):
        return (self.study.bound_spec(i) for i in self._input_conns)

    @property
    def outputs(self):
        return (self.study.bound_spec(o) for o in self._output_conns)

    @property
    def input_names(self):
        return self._input_conns.keys()

    @property
    def output_names(self):
        return self._output_conns.keys()

    @property
    def joins(self):
        return self._iterator_joins

    @property
    def joins_subjects(self):
        "Iterators that are joined within the pipeline"
        return self.study.SUBJECT_ID in self._iterator_joins

    @property
    def joins_visits(self):
        "Iterators that are joined within the pipeline"
        return self.study.VISIT_ID in self._iterator_joins

    @property
    def input_frequencies(self):
        freqs = set(i.frequency for i in self.inputs)
        if self.study.SUBJECT_ID in self._iterator_conns:
            freqs.add('per_subject')
        if self.study.VISIT_ID in self._iterator_conns:
            freqs.add('per_visit')
        return freqs

    @property
    def output_frequencies(self):
        return set(o.frequency for o in self.outputs)

    def frequency_inputs(self, frequency):
        return (i for i in self.inputs if i.frequency == frequency)

    def frequency_outputs(self, frequency):
        return (o for o in self.outputs if o.frequency == frequency)

    @property
    def all_parameters(self):
        """Return all parameters, including parameters of prerequisites"""
        return chain(self.parameters, iter(self._prereq_parameters.items()))

    @property
    def non_default_parameters(self):
        return ((k, v) for k, v in self.parameters.items()
                if v != self.default_parameters[k])

    @property
    def required_outputs(self):
        return (self.study.bound_spec(o) for o in self._required_outputs)

    @property
    def desc(self):
        return self._desc

    @classmethod
    def requires_conversion(cls, fileset, file_format):
        """Checks whether the fileset matches the requested file format"""
        if file_format is None:
            return False
        try:
            filset_format = fileset.format
        except AttributeError:
            return False  # Field input
        else:
            return (file_format != filset_format)

    def node(self, name):
        return self.workflow.get_node('{}_{}'.format(self.name, name))

    @property
    def nodes(self):
        return (self.workflow.get_node(n)
                for n in self.workflow.list_node_names())

    @property
    def node_names(self):
        return (n[len(self.name) + 1:]
                for n in self.workflow.list_node_names())

    def save_graph(self, fname, style='flat', format='png', **kwargs):  # @ReservedAssignment @IgnorePep8
        """
        Saves a graph of the pipeline to file

        Parameters
        ----------
        fname : str
            The filename for the saved graph
        style : str
            The style of the graph, can be one of can be one of
            'orig', 'flat', 'exec', 'hierarchical'
        plot : bool
            Whether to load and plot the graph after it has been written
        """
        fname = os.path.expanduser(fname)
        if not fname.endswith('.png'):
            fname += '.png'
        orig_dir = os.getcwd()
        tmpdir = tempfile.mkdtemp()
        os.chdir(tmpdir)
        workflow = self._workflow
        workflow.write_graph(graph2use=style, format=format, **kwargs)
        os.chdir(orig_dir)
        try:
            shutil.move(os.path.join(tmpdir, 'graph_detailed.{}'
                                     .format(format)), fname)
        except IOError as e:
            if e.errno == errno.ENOENT:
                shutil.move(os.path.join(tmpdir, 'graph.{}'.format(format)),
                            fname)
            else:
                raise
        shutil.rmtree(tmpdir)

    def iterators(self, frequency=None):
        """
        Returns the iterators (i.e. subject_id, visit_id) that the pipeline
        iterates over

        Parameters
        ----------
        frequency : str | None
            A selected data frequency to use to determine which iterators are
            required. If None, all input frequencies of the pipeline are
            assumed
        """
        iterators = set()
        if frequency is None:
            input_freqs = list(self.input_frequencies)
        else:
            input_freqs = [frequency]
        for freq in input_freqs:
            iterators.update(self.study.FREQUENCIES[freq])
        return iterators

    def iterates_over(self, iterator, freq):
        """
        Checks to see if the given frequency requires iteration over the
        given iterator

        Parameters
        ----------
        iterator : str
            The iterator to check
        freq : str
            The frequency to check
        """
        return iterator in self.study.FREQUENCIES[freq]

    def _unwrap_maps(self, name_maps, name, study=None, **inner_maps):
        """
        Unwraps potentially nested name-mapping dictionaries to get values
        for name, input_map, output_map and study. Unsed in __init__.

        Parameters
        ----------
        name_maps : dict
            A dictionary containing the name_maps to apply to the values
        name : str
            Name passed from inner pipeline constructor
        study : Study
            The study to bind the pipeline to. Will be overridden by any values
            in the mods dict
        inner_maps : dict[str, dict[str,str]]
            input and output maps from inner pipeline constructors

        Returns
        -------
        name : str
            Potentially modified name of the pipeline
        study : Study
            Potentially modified study
        maps : dict[str, dict[str,str]]
            Potentially modifed input and output maps
        """
        # Set values of name and study
        name = name_maps.get('name', name)
        name = name_maps.get('prefix', '') + name
        study = name_maps.get('study', study)
        # Flatten input and output maps, combining maps from inner nests with
        # those in the "mods" dictionary
        maps = {}
        for mtype in ('input_map', 'output_map'):
            try:
                inner_map = inner_maps[mtype]
            except KeyError:
                try:
                    maps[mtype] = name_maps[mtype]  # Only outer map
                except KeyError:
                    pass  # No maps
            else:
                try:
                    outer_map = name_maps[mtype]
                except KeyError:
                    maps[mtype] = inner_map  # Only inner map
                else:
                    # Work through different combinations of  inner and outer
                    # map types (i.e. str & str, str & dict, dict & str, and
                    # dict & dict) and combine into a single map
                    if isinstance(outer_map, basestring):
                        if isinstance(inner_map, basestring):
                            # Concatenate prefixes
                            maps[mtype] = outer_map + inner_map
                        elif isinstance(inner_map, dict):
                            # Add outer_map prefix to all values in inner map
                            # dictionary
                            maps[mtype] = {k: outer_map + v
                                           for k, v in inner_map.items()}
                        else:
                            raise ArcanaDesignError(
                                "Unrecognised type for name map in '{}' "
                                "pipeline can be str or dict[str,str]: {}"
                                .format(name, inner_map))
                    elif isinstance(outer_map, dict):
                        if isinstance(inner_map, basestring):
                            # Strip inner map prefix from outer dictionary
                            # (which should have prefix included). This should
                            # be an unlikely case I imagine
                            maps[mtype] = {k[len(inner_map):]: v
                                           for k, v in outer_map.items()}
                        elif isinstance(inner_map, dict):
                            # Chain outer_map dictionary to inner map
                            # dictionary
                            maps[mtype] = deepcopy(outer_map)
                            maps[mtype].update(
                                {k: outer_map.get(v, v)
                                 for k, v in inner_map.items()})
                        else:
                            raise ArcanaDesignError(
                                "Unrecognised type for name map in '{}' "
                                "pipeline can be str or dict[str,str]: {}"
                                .format(name, inner_map))
                    else:
                        raise ArcanaDesignError(
                            "Unrecognised type for name map in '{}' "
                            "pipeline can be str or dict[str,str]: {}"
                            .format(name, outer_map))
        try:
            outer_maps = name_maps['name_maps']
        except KeyError:
            pass
        else:
            name, study, maps = self._unwrap_maps(
                outer_maps, name=name, study=study, **maps)
        return name, study, maps

    @property
    def _error_msg_loc(self):
        return "'{}' pipeline in {} class".format(self.name,
                                                  type(self.study).__name__)

    def inputnode(self, frequency):
        """
        Returns the input node for the given frequency.

        Parameters
        ----------
        frequency : str
            The frequency (i.e. 'per_session', 'per_visit', 'per_subject' or
            'per_study') of the input node to retrieve

        Returns
        -------
        inputnode : arcana.environment.base.Node
            The input node corresponding to the given frequency
        """
        if self._inputnodes is None:
            raise ArcanaUsageError(
                "The pipeline must be capped (see cap() method) before an "
                "input node is accessed")
        return self._inputnodes[frequency]

    def outputnode(self, frequency):
        """
        Returns the output node for the given frequency.

        Parameters
        ----------
        frequency : str
            The frequency (i.e. 'per_session', 'per_visit', 'per_subject' or
            'per_study') of the output node to retrieve

        Returns
        -------
        outputnode : arcana.environment.base.Node
            The output node corresponding to the given frequency
        """
        if self._outputnodes is None:
            raise ArcanaUsageError(
                "The pipeline must be capped (see cap() method) before an "
                "output node is accessed")
        return self._outputnodes[frequency]

    @property
    def prov(self):
        if self._prov is None:
            raise ArcanaUsageError(
                "The pipeline must be capped (see cap() method) before the "
                "provenance is accessed")
        return self._prov

    def cap(self):
        """
        "Caps" the construction of the pipeline, signifying that no more inputs
        and outputs are expected to be added and therefore the input and output
        nodes can be created along with the provenance.
        """
        to_cap = (self._inputnodes, self._outputnodes, self._prov)
        if to_cap == (None, None, None):
            self._inputnodes = {
                f: self._make_inputnode(f) for f in self.input_frequencies}
            self._outputnodes = {
                f: self._make_outputnode(f) for f in self.output_frequencies}
            self._prov = self._gen_prov()
        elif None in to_cap:
            raise ArcanaError(
                "If one of _inputnodes, _outputnodes or _prov is not None then"
                " they all should be in {}".format(self))

    def _make_inputnode(self, frequency):
        """
        Generates an input node for the given frequency. It also adds implicit
        file format conversion nodes to the pipeline.

        Parameters
        ----------
        frequency : str
            The frequency (i.e. 'per_session', 'per_visit', 'per_subject' or
            'per_study') of the input node to retrieve
        """
        # Check to see whether there are any outputs for the given frequency
        inputs = list(self.frequency_inputs(frequency))
        # Get list of input names for the requested frequency, addding fields
        # to hold iterator IDs
        input_names = [i.name for i in inputs]
        input_names.extend(self.study.FREQUENCIES[frequency])
        if not input_names:
            raise ArcanaError(
                "No inputs to '{}' pipeline for requested freqency '{}'"
                .format(self.name, frequency))
        # Generate input node and connect it to appropriate nodes
        inputnode = self.add('{}_inputnode'.format(frequency),
                             IdentityInterface(fields=input_names))
        # Loop through list of nodes connected to study data specs and
        # connect them to the newly created input node
        for input in inputs:  # @ReservedAssignment
            # Keep track of previous conversion nodes to avoid replicating the
            # conversion for inputs that are used in multiple places
            prev_conv_nodes = {}
            for (node, node_in, format,  # @ReservedAssignment @IgnorePep8
                 conv_kwargs) in self._input_conns[input.name]:
                # If fileset formats differ between study and pipeline
                # inputs create converter node (if one hasn't been already)
                # and connect input to that before connecting to inputnode
                if self.requires_conversion(input, format):
                    try:
                        conv = format.converter_from(input.format,
                                                     **conv_kwargs)
                    except ArcanaNoConverterError as e:
                        e.msg += (
                            "which is required to convert '{}' from {} to {} "
                            "for '{}' input of '{}' node".format(
                                input.name, input.format, format, node_in,
                                node.name))
                        raise e
                    try:
                        in_node = prev_conv_nodes[format.name]
                    except KeyError:
                        in_node = prev_conv_nodes[format.name] = self.add(
                            'conv_{}_to_{}_format'.format(input.name,
                                                          format.name),
                            conv.interface,
                            inputs={conv.input: (inputnode, input.name)},
                            requirements=conv.requirements,
                            mem_gb=conv.mem_gb,
                            wall_time=conv.wall_time)
                    in_node_out = conv.output
                else:
                    in_node = inputnode
                    in_node_out = input.name
                self.connect(in_node, in_node_out, node, node_in)
        # Connect iterator inputs
        for iterator, conns in self._iterator_conns.items():
            # Check to see if this is the right frequency for the iterator
            # input, i.e. if it is the only iterator for this frequency
            if self.study.FREQUENCIES[frequency] == (iterator,):
                for (node, node_in, format) in conns:  # @ReservedAssignment
                    self.connect(inputnode, iterator, node, node_in)
        return inputnode

    def _make_outputnode(self, frequency):
        """
        Generates an output node for the given frequency. It also adds implicit
        file format conversion nodes to the pipeline.

        Parameters
        ----------
        frequency : str
            The frequency (i.e. 'per_session', 'per_visit', 'per_subject' or
            'per_study') of the output node to retrieve
        """
        # Check to see whether there are any outputs for the given frequency
        outputs = list(self.frequency_outputs(frequency))
        if not outputs:
            raise ArcanaError(
                "No outputs to '{}' pipeline for requested freqency '{}'"
                .format(self.name, frequency))
        # Get list of output names for the requested frequency, addding fields
        # to hold iterator IDs
        output_names = [o.name for o in outputs]
        # Generate output node and connect it to appropriate nodes
        outputnode = self.add('{}_outputnode'.format(frequency),
                              IdentityInterface(fields=output_names))
        # Loop through list of nodes connected to study data specs and
        # connect them to the newly created output node
        for output in outputs:  # @ReservedAssignment
            (node, node_out, format,  # @ReservedAssignment @IgnorePep8
             conv_kwargs) = self._output_conns[output.name]
            # If fileset formats differ between study and pipeline
            # outputs create converter node (if one hasn't been already)
            # and connect output to that before connecting to outputnode
            if self.requires_conversion(output, format):
                conv = output.format.converter_from(format, **conv_kwargs)
                node = self.add(
                    'conv_{}_from_{}_format'.format(output.name, format.name),
                    conv.interface,
                    inputs={conv.input: (node, node_out)},
                    requirements=conv.requirements,
                    mem_gb=conv.mem_gb,
                    wall_time=conv.wall_time)
                node_out = conv.output
            self.connect(node, node_out, outputnode, output.name)
        return outputnode

    def _gen_prov(self):
        """
        Extracts provenance information from the pipeline into a PipelineProv
        object

        Returns
        -------
        prov : dict[str, *]
            A dictionary containing the provenance information to record
            for the pipeline
        """
        # Export worfklow graph to node-link data format
        wf_dict = nx_json.node_link_data(self.workflow._graph)
        # Replace references to Node objects with the node's provenance
        # information and convert to a dict organised by node name to allow it
        # to be compared more easily. Also change link node-references from
        # node index to node ID so it is not dependent on the order the nodes
        # are written to the dictionary (which for Python < 3.7 is guaranteed
        # to be the same between identical runs)
        for link in wf_dict['links']:
            if int(networkx_version.split('.')[0]) < 2:  # @UndefinedVariable
                link['source'] = wf_dict['nodes'][link['source']]['id'].name
                link['target'] = wf_dict['nodes'][link['target']]['id'].name
            else:
                link['source'] = link['source'].name
                link['target'] = link['target'].name
        wf_dict['nodes'] = {n['id'].name: n['id'].prov
                            for n in wf_dict['nodes']}
        # Roundtrip to JSON to convert any tuples into lists so dictionaries
        # can be compared directly
        wf_dict = json.loads(json.dumps(wf_dict))
        dependency_versions = {d: extract_package_version(d)
                               for d in ARCANA_DEPENDENCIES}
        pkg_versions = {'arcana': __version__}
        pkg_versions.update((k, v) for k, v in dependency_versions.items()
                            if v is not None)
        prov = {
            '__prov_version__': PROVENANCE_VERSION,
            'name': self.name,
            'workflow': wf_dict,
            'study': self.study.prov,
            'pkg_versions': pkg_versions,
            'python_version': sys.version,
            'joined_ids': self._joined_ids()}
        return prov

    def expected_record(self, node):
        """
        Constructs the provenance record that would be saved in the given node
        if the pipeline was run on the current state of the repository

        Parameters
        ----------
        node : arcana.repository.tree.TreeNode
            A node of the Tree representation of the study data stored in the
            repository (i.e. a Session, Visit, Subject or Tree node)

        Returns
        -------
        expected_record : arcana.provenance.Record
            The record that would be produced if the pipeline is run over the
            study tree.
        """
        exp_inputs = {}
        # Get checksums/values of all inputs that would have been used in
        # previous runs of an equivalent pipeline to compare with that saved
        # in provenance to see if any have been updated.
        for inpt in self.inputs:  # @ReservedAssignment
            # Get iterators present in the input that aren't in this node
            # and need to be joined
            iterators_to_join = (self.iterators(inpt.frequency) -
                                 self.iterators(node.frequency))
            if not iterators_to_join:
                # No iterators to join so we can just extract the checksums
                # of the corresponding input
                exp_inputs[inpt.name] = inpt.collection.item(
                    node.subject_id, node.visit_id).checksums
            elif len(iterators_to_join) == 1:
                # Get list of checksums dicts for each node of the input
                # frequency that relates to the current node
                exp_inputs[inpt.name] = [
                    inpt.collection.item(n.subject_id, n.visit_id).checksums
                    for n in node.nodes(inpt.frequency)]
            else:
                # In the case where the node is the whole treee and the input
                # is per_seession, we need to create a list of lists to match
                # how the checksums are joined in the processor
                exp_inputs[inpt.name] = []
                for subj in node.subjects:
                    exp_inputs[inpt.name].append([
                        inpt.collection.item(s.subject_id,
                                             s.visit_id).checksums
                        for s in subj.sessions])
        # Get checksums/value for all outputs of the pipeline. We are assuming
        # that they exist here (otherwise they will be None)
        exp_outputs = {
            o.name: o.collection.item(node.subject_id, node.visit_id).checksums
            for o in self.outputs}
        exp_prov = copy(self.prov)
        if PY2:
            # Need to convert to unicode strings for Python 2
            exp_inputs = json.loads(json.dumps(exp_inputs))
            exp_outputs = json.loads(json.dumps(exp_outputs))
        exp_prov['inputs'] = exp_inputs
        exp_prov['outputs'] = exp_outputs
        exp_prov['joined_ids'] = self._joined_ids()
        return Record(
            self.name, node.frequency, node.subject_id, node.visit_id,
            self.study.name, exp_prov)

    def _joined_ids(self):
        """
        Adds the subjects/visits used to generate the derivatives iff there are
        any joins over them in the pipeline
        """
        joined_prov = {}
        if self.joins_subjects:
            joined_prov['subject_ids'] = list(self.study.subject_ids)
        if self.joins_visits:
            joined_prov['visit_ids'] = list(self.study.visit_ids)
        return joined_prov



from __future__ import absolute_import
from builtins import range
from builtins import object
import string
import random
import os
import os.path as op
import shutil
import re
from copy import copy
import pydicom
import xnat
from arcana.utils.testing import BaseTestCase
from arcana.repository.xnat import XnatRepo
from arcana.repository.basic import BasicRepo
from arcana.study import Study, StudyMetaClass
from arcana.data import InputFilesetSpec, FilesetSpec, FieldSpec
from arcana.exceptions import ArcanaError, ArcanaFileFormatError
from arcana.data.file_format import text_format
import logging
from future.utils import with_metaclass


logger = logging.getLogger('arcana')

try:
    SERVER = os.environ['ARCANA_TEST_XNAT']
except KeyError:
    SERVER = None

SKIP_ARGS = (SERVER is None, "Skipping as ARCANA_TEST_XNAT env var not set")


class CreateXnatProjectMixin(object):

    PROJECT_NAME_LEN = 12
    REF_FORMATS = [text_format]

    @property
    def project(self):
        """
        Creates a random string of letters and numbers to be the
        project ID
        """
        try:
            return self._project
        except AttributeError:
            self._project = ''.join(
                random.choice(string.ascii_uppercase + string.digits)
                for _ in range(self.PROJECT_NAME_LEN))
            return self._project

    def _create_project(self, project_name=None):
        if project_name is None:
            project_name = self.project
        if SERVER == 'https://mbi-xnat.erc.monash.edu.au':
            raise ArcanaError(
                "Shouldn't be creating projects on the production "
                "server")
        with xnat.connect(SERVER) as login:
            uri = '/data/archive/projects/{}'.format(project_name)
            query = {'xsiType': 'xnat:projectData', 'req_format': 'qa'}
            response = login.put(uri, query=query)
            if response.ok:
                logger.info("Created test project '{}'"
                            .format(project_name))

    def _delete_project(self, project_name=None):
        if project_name is None:
            project_name = self.project
        with xnat.connect(SERVER) as login:
            login.projects[project_name].delete()


class TestOnXnatMixin(CreateXnatProjectMixin):

    def session_label(self, project=None, subject=None, visit=None,
                      from_study=None):
        if project is None:
            project = self.project
        if subject is None:
            subject = self.SUBJECT
        if visit is None:
            visit = self.VISIT
        label = '_'.join((project, subject, visit))
        if from_study is not None:
            label += '_' + from_study
        return label

    def session_cache(self, base_dir=None, project=None, subject=None,
                      visit=None, from_study=None):
        if base_dir is None:
            base_dir = self.cache_dir
        if project is None:
            project = self.project
        if subject is None:
            subject = self.SUBJECT
        return op.join(
            base_dir, project, '{}_{}'.format(project, subject),
            self.session_label(project=project, subject=subject,
                               visit=visit, from_study=from_study))

    def setUp(self):
        BaseTestCase.setUp(self)
        shutil.rmtree(self.cache_dir, ignore_errors=True)
        os.makedirs(self.cache_dir)
        self._create_project()
        with self._connect() as login:
            xproject = login.projects[self.project]
            xsubject = login.classes.SubjectData(
                label='{}_{}'.format(self.project, self.SUBJECT),
                parent=xproject)
            xsession = login.classes.MrSessionData(
                label=self.session_label(),
                parent=xsubject)
            for fileset in self.session.filesets:
                fileset.format = fileset.detect_format(self.REF_FORMATS)
                if fileset.format.name == 'dicom':
                    dcm_files = [f for f in os.listdir(fileset.path)
                                 if f.endswith('.dcm')]
                    hdr = pydicom.dcmread(op.join(fileset.path, dcm_files[0]))
                    id_ = int(hdr.SeriesNumber)
                else:
                    id_ = fileset.basename
                xfileset = login.classes.MrScanData(id=id_,
                                                    type=fileset.basename,
                                                    parent=xsession)
                resource = xfileset.create_resource(
                    fileset.format.resource_names(XnatRepo.type)[0])
                if fileset.format.directory:
                    for fname in os.listdir(fileset.path):
                        resource.upload(
                            op.join(fileset.path, fname), fname)
                else:
                    for path in fileset.paths:
                        resource.upload(path, op.basename(path))
            for field in self.session.fields:
                if field.dtype is str:
                    value = '"{}"'.format(field.value)
                else:
                    value = field.value
                xsession.fields[field.name] = value

    def tearDown(self):
        # Clean up working dirs
        shutil.rmtree(self.cache_dir, ignore_errors=True)
        # Clean up session created for unit-test
        self._delete_project()

    def _connect(self):
        return xnat.connect(SERVER)


class TestMultiSubjectOnXnatMixin(CreateXnatProjectMixin):

    sanitize_id_re = re.compile(r'[^a-zA-Z_0-9]')

    def setUp(self):
        self._clean_up()
        self._repository = XnatRepo(project_id=self.project,
                                          server=SERVER,
                                          cache_dir=self.cache_dir)
        self._create_project()
        self.BASE_CLASS.setUp(self)
        local_repository = BasicRepo(self.project_dir)
        tree = local_repository.tree()
        repo = XnatRepo(SERVER, self.project, '/tmp')
        with repo:
            for node in tree:
                for fileset in node.filesets:
                    # Need to forcibly change the repository to be XNAT
                    fileset = copy(fileset)
                    fileset.format = fileset.detect_format(self.REF_FORMATS)
                    fileset._repository = repo
                    fileset.put()
                for field in node.fields:
                    # Need to forcibly change the repository to be XNAT
                    field = copy(field)
                    field._repository = repo
                    field.put()
                for record in node.records:
                    repo.put_record(record)

    def tearDown(self):
        self._clean_up()
        self._delete_project()

    def _clean_up(self):
        # Clean up working dirs
        shutil.rmtree(self.cache_dir, ignore_errors=True)

    @property
    def repository(self):
        return self._repository

    @property
    def xnat_session_name(self):
        return '{}_{}'.format(self.project, self.base_name)

    @property
    def project_dir(self):
        return op.join(self.repository_path, self.base_name)

    @property
    def output_cache_dir(self):
        return self._output_cache_dir

    @property
    def base_name(self):
        return self.BASE_CLASS._get_name()

    def _full_subject_id(self, subject):
        return self.project + '_' + subject

    def get_session_dir(self, subject=None, visit=None,
                        frequency='per_session'):
        if subject is None and frequency in ('per_session', 'per_subject'):
            subject = self.SUBJECT
        if visit is None and frequency in ('per_session', 'per_visit'):
            visit = self.VISIT
        if frequency == 'per_session':
            assert subject is not None
            assert visit is not None
            parts = [self.project, subject, visit]
        elif frequency == 'per_subject':
            assert subject is not None
            assert visit is None
            parts = [self.project, subject, XnatRepo.SUMMARY_NAME]
        elif frequency == 'per_visit':
            assert visit is not None
            assert subject is None
            parts = [self.project, XnatRepo.SUMMARY_NAME, visit]
        elif frequency == 'per_study':
            assert subject is None
            assert visit is None
            parts = [self.project, XnatRepo.SUMMARY_NAME,
                     XnatRepo.SUMMARY_NAME]
        else:
            assert False
        session_id = '_'.join(parts)
        session_path = op.join(self.output_cache_dir, session_id)
        if not op.exists(session_path):
            raise ArcanaError(
                "Session path '{}' does not exist".format(session_path))
        return session_path

    def output_file_path(self, fname, from_study, subject=None, visit=None,
                         frequency='per_session'):
        try:
            acq_path = self.BASE_CLASS.output_file_path(
                self, fname, from_study, subject=subject, visit=visit,
                frequency=frequency, derived=False)
        except KeyError:
            acq_path = None
        try:
            proc_path = self.BASE_CLASS.output_file_path(
                self, fname, from_study, subject=subject, visit=visit,
                frequency=frequency, derived=True)
        except KeyError:
            proc_path = None
        if acq_path is not None and op.exists(acq_path):
            if op.exists(proc_path):
                raise ArcanaError(
                    "Both acquired and derived paths were found for "
                    "'{}_{}' ({} and {})".format(from_study, fname, acq_path,
                                                 proc_path))
            path = acq_path
        else:
            path = proc_path
        return path


def filter_scans(names):
    return sorted(f for f in sorted(names)
                  if (f != XnatRepo.PROV_SCAN and
                      not f.endswith(XnatRepo.MD5_SUFFIX)))



from arcana.utils.testing import BaseTestCase, TestMath
from arcana.utils.interfaces import Merge
from arcana.data import InputFileset, FilesetSpec, InputFilesetSpec
from arcana.data.file_format import text_format
from arcana.study.parameter import ParamSpec
from arcana.exceptions import ArcanaOutputNotProducedException
from arcana.study.base import Study
from arcana.study.multi import (
    MultiStudy, SubStudySpec, MultiStudyMetaClass, StudyMetaClass)
from arcana.study.parameter import Parameter
from future.utils import with_metaclass
import unittest


class NotSpecifiedRequiredParameter(Exception):
    pass


class StudyA(with_metaclass(StudyMetaClass, Study)):

    add_data_specs = [
        InputFilesetSpec('x', text_format),
        InputFilesetSpec('y', text_format),
        FilesetSpec('z', text_format, 'pipeline_alpha')]

    add_param_specs = [
        ParamSpec('o1', 1),
        ParamSpec('o2', '2'),
        ParamSpec('o3', 3.0)]

    def pipeline_alpha(self, **name_maps):  # @UnusedVariable
        pipeline = self.new_pipeline(
            name='pipeline_alpha',
            desc="A dummy pipeline used to test MultiStudy class",
            citations=[],
            name_maps=name_maps)
        math = pipeline.add("math", TestMath())
        math.inputs.op = 'add'
        math.inputs.as_file = True
        # Connect inputs
        pipeline.connect_input('x', math, 'x')
        pipeline.connect_input('y', math, 'y')
        # Connect outputs
        pipeline.connect_output('z', math, 'z')
        return pipeline


class StudyB(with_metaclass(StudyMetaClass, Study)):

    add_data_specs = [
        InputFilesetSpec('w', text_format),
        InputFilesetSpec('x', text_format),
        FilesetSpec('y', text_format, 'pipeline_beta'),
        FilesetSpec('z', text_format, 'pipeline_beta')]

    add_param_specs = [
        ParamSpec('o1', 10),
        ParamSpec('o2', '20'),
        ParamSpec('o3', 30.0),
        ParamSpec('product_op', None, dtype=str)]  # Needs to be set to 'product' @IgnorePep8

    def pipeline_beta(self, **name_maps):  # @UnusedVariable
        pipeline = self.new_pipeline(
            name='pipeline_beta',
            desc="A dummy pipeline used to test MultiStudy class",
            citations=[],
            name_maps=name_maps)
        add1 = pipeline.add("add1", TestMath())
        add2 = pipeline.add("add2", TestMath())
        prod = pipeline.add("product", TestMath())
        add1.inputs.op = 'add'
        add2.inputs.op = 'add'
        if self.parameter('product_op') is None:
            raise NotSpecifiedRequiredParameter
        prod.inputs.op = self.parameter('product_op')
        add1.inputs.as_file = True
        add2.inputs.as_file = True
        prod.inputs.as_file = True
        # Connect inputs
        pipeline.connect_input('w', add1, 'x')
        pipeline.connect_input('x', add1, 'y')
        pipeline.connect_input('x', add2, 'x')
        # Connect nodes
        pipeline.connect(add1, 'z', add2, 'y')
        pipeline.connect(add1, 'z', prod, 'x')
        pipeline.connect(add2, 'z', prod, 'y')
        # Connect outputs
        pipeline.connect_output('y', add2, 'z')
        pipeline.connect_output('z', prod, 'z')
        return pipeline


class FullMultiStudy(with_metaclass(MultiStudyMetaClass, MultiStudy)):

    add_substudy_specs = [
        SubStudySpec('ss1', StudyA,
                     {'x': 'a',
                      'y': 'b',
                      'z': 'd',
                      'o1': 'p1',
                      'o2': 'p2',
                      'o3': 'p3'}),
        SubStudySpec('ss2', StudyB,
                     {'w': 'b',
                      'x': 'c',
                      'y': 'e',
                      'z': 'f',
                      'o1': 'q1',
                      'o2': 'q2',
                      'o3': 'p3',
                      'product_op': 'required_op'})]

    add_data_specs = [
        InputFilesetSpec('a', text_format),
        InputFilesetSpec('b', text_format),
        InputFilesetSpec('c', text_format),
        FilesetSpec('d', text_format, 'pipeline_alpha_trans'),
        FilesetSpec('e', text_format, 'pipeline_beta_trans'),
        FilesetSpec('f', text_format, 'pipeline_beta_trans')]

    add_param_specs = [
        ParamSpec('p1', 100),
        ParamSpec('p2', '200'),
        ParamSpec('p3', 300.0),
        ParamSpec('q1', 150),
        ParamSpec('q2', '250'),
        ParamSpec('required_op', None, dtype=str)]

    pipeline_alpha_trans = MultiStudy.translate(
        'ss1', 'pipeline_alpha')
    pipeline_beta_trans = MultiStudy.translate(
        'ss2', 'pipeline_beta')


class PartialMultiStudy(with_metaclass(MultiStudyMetaClass, MultiStudy)):

    add_substudy_specs = [
        SubStudySpec('ss1', StudyA,
                     {'x': 'a', 'y': 'b', 'o1': 'p1'}),
        SubStudySpec('ss2', StudyB,
                     {'w': 'b', 'x': 'c', 'o1': 'p1'})]

    add_data_specs = [
        InputFilesetSpec('a', text_format),
        InputFilesetSpec('b', text_format),
        InputFilesetSpec('c', text_format)]

    pipeline_alpha_trans = MultiStudy.translate(
        'ss1', 'pipeline_alpha')

    add_param_specs = [
        ParamSpec('p1', 1000)]


class MultiMultiStudy(with_metaclass(MultiStudyMetaClass, MultiStudy)):

    add_substudy_specs = [
        SubStudySpec('ss1', StudyA),
        SubStudySpec('full', FullMultiStudy),
        SubStudySpec('partial', PartialMultiStudy)]

    add_data_specs = [
        FilesetSpec('g', text_format, 'combined_pipeline')]

    add_param_specs = [
        ParamSpec('combined_op', 'add')]

    def combined_pipeline(self, **name_maps):
        pipeline = self.new_pipeline(
            name='combined',
            desc=(
                "A dummy pipeline used to test MultiMultiStudy class"),
            citations=[],
            name_maps=name_maps)
        merge = pipeline.add("merge", Merge(3))
        math = pipeline.add("math", TestMath())
        math.inputs.op = self.parameter('combined_op')
        math.inputs.as_file = True
        # Connect inputs
        pipeline.connect_input('ss1_z', merge, 'in1')
        pipeline.connect_input('full_e', merge, 'in2')
        pipeline.connect_input('partial_ss2_z', merge, 'in3')
        # Connect nodes
        pipeline.connect(merge, 'out', math, 'x')
        # Connect outputs
        pipeline.connect_output('g', math, 'z')
        return pipeline


class TestMulti(BaseTestCase):

    INPUT_FILESETS = {'ones': '1'}

    def test_full_multi_study(self):
        study = self.create_study(
            FullMultiStudy, 'full',
            [InputFileset('a', 'ones', text_format),
             InputFileset('b', 'ones', text_format),
             InputFileset('c', 'ones', text_format)],
            parameters=[Parameter('required_op', 'mul')])
        d, e, f = study.data(('d', 'e', 'f'),
                             subject_id='SUBJECT', visit_id='VISIT')
        self.assertContentsEqual(d, 2.0)
        self.assertContentsEqual(e, 3.0)
        self.assertContentsEqual(f, 6.0)
        # Test parameter values in MultiStudy
        self.assertEqual(study._get_parameter('p1').value, 100)
        self.assertEqual(study._get_parameter('p2').value, '200')
        self.assertEqual(study._get_parameter('p3').value, 300.0)
        self.assertEqual(study._get_parameter('q1').value, 150)
        self.assertEqual(study._get_parameter('q2').value, '250')
        self.assertEqual(study._get_parameter('required_op').value, 'mul')
        # Test parameter values in SubStudy
        ss1 = study.substudy('ss1')
        self.assertEqual(ss1._get_parameter('o1').value, 100)
        self.assertEqual(ss1._get_parameter('o2').value, '200')
        self.assertEqual(ss1._get_parameter('o3').value, 300.0)
        ss2 = study.substudy('ss2')
        self.assertEqual(ss2._get_parameter('o1').value, 150)
        self.assertEqual(ss2._get_parameter('o2').value, '250')
        self.assertEqual(ss2._get_parameter('o3').value, 300.0)
        self.assertEqual(ss2._get_parameter('product_op').value, 'mul')

    def test_partial_multi_study(self):
        study = self.create_study(
            PartialMultiStudy, 'partial',
            [InputFileset('a', 'ones', text_format),
             InputFileset('b', 'ones', text_format),
             InputFileset('c', 'ones', text_format)],
            parameters=[Parameter('ss2_product_op', 'mul')])
        ss1_z = study.data('ss1_z',
                           subject_id='SUBJECT', visit_id='VISIT')
        ss2_z = list(study.data('ss2_z'))[0]
        self.assertContentsEqual(ss1_z, 2.0)
        self.assertContentsEqual(study.data('ss2_y'), 3.0)
        self.assertContentsEqual(ss2_z, 6.0)
        # Test parameter values in MultiStudy
        self.assertEqual(study._get_parameter('p1').value, 1000)
        self.assertEqual(study._get_parameter('ss1_o2').value, '2')
        self.assertEqual(study._get_parameter('ss1_o3').value, 3.0)
        self.assertEqual(study._get_parameter('ss2_o2').value, '20')
        self.assertEqual(study._get_parameter('ss2_o3').value, 30.0)
        self.assertEqual(study._get_parameter('ss2_product_op').value, 'mul')
        # Test parameter values in SubStudy
        ss1 = study.substudy('ss1')
        self.assertEqual(ss1._get_parameter('o1').value, 1000)
        self.assertEqual(ss1._get_parameter('o2').value, '2')
        self.assertEqual(ss1._get_parameter('o3').value, 3.0)
        ss2 = study.substudy('ss2')
        self.assertEqual(ss2._get_parameter('o1').value, 1000)
        self.assertEqual(ss2._get_parameter('o2').value, '20')
        self.assertEqual(ss2._get_parameter('o3').value, 30.0)
        self.assertEqual(ss2._get_parameter('product_op').value, 'mul')

    def test_multi_multi_study(self):
        study = self.create_study(
            MultiMultiStudy, 'multi_multi',
            [InputFileset('ss1_x', 'ones', text_format),
             InputFileset('ss1_y', 'ones', text_format),
             InputFileset('full_a', 'ones', text_format),
             InputFileset('full_b', 'ones', text_format),
             InputFileset('full_c', 'ones', text_format),
             InputFileset('partial_a', 'ones', text_format),
             InputFileset('partial_b', 'ones', text_format),
             InputFileset('partial_c', 'ones', text_format)],
            parameters=[Parameter('full_required_op', 'mul'),
                        Parameter('partial_ss2_product_op', 'mul')])
        self.assertContentsEqual(study.data('g'), 11.0)
        # Test parameter values in MultiStudy
        self.assertEqual(study._get_parameter('full_p1').value, 100)
        self.assertEqual(study._get_parameter('full_p2').value, '200')
        self.assertEqual(study._get_parameter('full_p3').value, 300.0)
        self.assertEqual(study._get_parameter('full_q1').value, 150)
        self.assertEqual(study._get_parameter('full_q2').value, '250')
        self.assertEqual(study._get_parameter('full_required_op').value,
                         'mul')
        # Test parameter values in SubStudy
        ss1 = study.substudy('full').substudy('ss1')
        self.assertEqual(ss1._get_parameter('o1').value, 100)
        self.assertEqual(ss1._get_parameter('o2').value, '200')
        self.assertEqual(ss1._get_parameter('o3').value, 300.0)
        ss2 = study.substudy('full').substudy('ss2')
        self.assertEqual(ss2._get_parameter('o1').value, 150)
        self.assertEqual(ss2._get_parameter('o2').value, '250')
        self.assertEqual(ss2._get_parameter('o3').value, 300.0)
        self.assertEqual(ss2._get_parameter('product_op').value, 'mul')
        # Test parameter values in MultiStudy
        self.assertEqual(study._get_parameter('partial_p1').value, 1000)
        self.assertEqual(study._get_parameter('partial_ss1_o2').value, '2')
        self.assertEqual(study._get_parameter('partial_ss1_o3').value, 3.0)
        self.assertEqual(study._get_parameter('partial_ss2_o2').value, '20')
        self.assertEqual(study._get_parameter('partial_ss2_o3').value, 30.0)
        self.assertEqual(
            study._get_parameter('partial_ss2_product_op').value, 'mul')
        # Test parameter values in SubStudy
        ss1 = study.substudy('partial').substudy('ss1')
        self.assertEqual(ss1._get_parameter('o1').value, 1000)
        self.assertEqual(ss1._get_parameter('o2').value, '2')
        self.assertEqual(ss1._get_parameter('o3').value, 3.0)
        ss2 = study.substudy('partial').substudy('ss2')
        self.assertEqual(ss2._get_parameter('o1').value, 1000)
        self.assertEqual(ss2._get_parameter('o2').value, '20')
        self.assertEqual(ss2._get_parameter('o3').value, 30.0)
        self.assertEqual(ss2._get_parameter('product_op').value, 'mul')

    def test_missing_parameter(self):
        # Misses the required 'full_required_op' parameter, which sets
        # the operation of the second node in StudyB's pipeline to
        # 'product'
        inputs = [InputFileset('ss1_x', 'ones', text_format),
                  InputFileset('ss1_y', 'ones', text_format),
                  InputFileset('full_a', 'ones', text_format),
                  InputFileset('full_b', 'ones', text_format),
                  InputFileset('full_c', 'ones', text_format),
                  InputFileset('partial_a', 'ones', text_format),
                  InputFileset('partial_b', 'ones', text_format),
                  InputFileset('partial_c', 'ones', text_format)]
        missing_parameter_study = self.create_study(
            MultiMultiStudy, 'multi_multi',
            inputs,
            parameters=[
                Parameter('partial_ss2_product_op', 'mul')])
        self.assertRaises(
            NotSpecifiedRequiredParameter,
            missing_parameter_study.data,
            'g')
        missing_parameter_study2 = self.create_study(
            MultiMultiStudy, 'multi_multi',
            inputs,
            parameters=[Parameter('full_required_op', 'mul')])
        self.assertRaises(
            NotSpecifiedRequiredParameter,
            missing_parameter_study2.data,
            'g')
        provided_parameters_study = self.create_study(
            MultiMultiStudy, 'multi_multi',
            inputs,
            parameters=[
                Parameter('partial_ss2_product_op', 'mul'),
                Parameter('full_required_op', 'mul')])
        g = list(provided_parameters_study.data('g'))[0]
        self.assertContentsEqual(g, 11.0)



from past.builtins import basestring
from builtins import object
from copy import copy
from arcana.exceptions import ArcanaUsageError


class Parameter(object):
    """
    Represents a parameter passed to a Study object

    Parameters
    ----------
    name : str
        Name of the parameter
    value : float | int | str | list | tuple
        Value of the parameter
    """

    def __init__(self, name, value):
        self._name = name
        if value is None:
            self._dtype = None
        else:
            if not isinstance(value, (int, float, basestring,
                                      tuple, list)):
                raise ArcanaUsageError(
                    "Invalid type for '{}' parameter default ({}), {}, "
                    "can be one of int, float or str"
                    .format(name, value, type(value)))
            self._dtype = (str
                           if isinstance(value, basestring) else type(value))
        self._value = value

    @property
    def name(self):
        return self._name

    @property
    def value(self):
        return self._value

    @property
    def dtype(self):
        if self._dtype is None:
            return type(None)
        return self._dtype

    def renamed(self, name):
        """
        Duplicate the Parameter and rename it
        """
        duplicate = copy(self)
        duplicate._name = name
        return duplicate

    def __repr__(self):
        return "Parameter(name='{}', value={})".format(self.name,
                                                       self.value)


class ParamSpec(Parameter):
    """
    Specifies a parameter that can be passed to the study

    Parameters
    ----------
    name : str
        Name of the parameter
    default : float | int | str | list | tuple
        Default value of the parameter
    choices : List(float | int | str | list | tuple)
        Restrict valid inputs to the following choices
    desc : str
        A description of the parameter
    dtype : type | None
        The datatype of the parameter. If none will be determined from
        default value
    """

    def __init__(self, name, default, desc=None, dtype=None,
                 array=False):
        super(ParamSpec, self).__init__(name, default)
        self._desc = desc
        self._array = array
        if dtype is not None:
            if self.default is not None and (
                not array and not isinstance(self.default, dtype) or
                array and any(not isinstance(d, dtype)
                              for d in self.default)):
                raise ArcanaUsageError(
                    "Provided default value ({}) does not match explicit "
                    "dtype ({})".format(self.default, dtype))
            self._dtype = dtype

    @property
    def name(self):
        return self._name

    @property
    def array(self):
        return self._array

    @property
    def default(self):
        return self._value

    @property
    def desc(self):
        return self._desc

    def __repr__(self):
        return "ParamSpec(name='{}', default={}, desc='{}')".format(
            self.name, self.default, self.desc)

    def check_valid(self, parameter, context=None):
        if parameter.value is not None:
            error_msg = (
                "Incorrect datatype for '{}' parameter provided "
                "({}){}, Should be {}"
                .format(parameter.name, type(parameter.value),
                        'in ' + context if context is not None else '',
                        self.dtype))
            if self.array:
                if any(not isinstance(v, self.dtype) for v in parameter.value):
                    raise ArcanaUsageError(error_msg + ' array')
            elif not isinstance(parameter.value, self.dtype):
                raise ArcanaUsageError(error_msg)


class SwitchSpec(ParamSpec):
    """
    Specifies a special parameter that switches between different
    methods and/or pipeline input/outputs. Typically used to select
    between comparable methods (e.g. FSL or ANTs registration) but can
    also be used to specify whether certain methods are applied, and by
    extension some auxiliary outputs are generated

    Parameters
    ----------
    name : str
        Name of the parameter
    default : str
        Default option for the switch
    choices : list[str]
        The valid values for the switch
    desc : str
        A description of the parameter
    """

    def __init__(self, name, default, choices=None, desc=None,
                 dtype=None):
        super(SwitchSpec, self).__init__(name, default, desc=desc,
                                         dtype=dtype)
        if self.is_boolean:
            if choices is not None:
                raise ArcanaUsageError(
                    "Choices ({}) are only valid for non-boolean "
                    "switches ('{}')".format("', '".join(choices),
                                               name))
        elif choices is None:
            raise ArcanaUsageError(
                "Choices must be provided for non-boolean "
                "switches ('{}')".format(name))
        self._choices = tuple(choices) if choices is not None else None
        self._desc = desc

    @property
    def name(self):
        return self._name

    @property
    def default(self):
        return self._value

    @property
    def is_boolean(self):
        return isinstance(self.default, bool)

    @property
    def choices(self):
        return self._choices

    def check_valid(self, switch, context=''):
        super(SwitchSpec, self).check_valid(switch, context=context)
        if self.is_boolean:
            if not isinstance(switch.value, bool):
                raise ArcanaUsageError(
                    "Value provided to switch '{}'{} should be a "
                    "boolean (not {})".format(
                        self.name, context, switch.value))
        elif switch.value not in self.choices:
            raise ArcanaUsageError(
                "Value provided to switch '{}'{} ({}) is not a valid "
                "choice ('{}')".format(
                    self.name, context, switch.value,
                    "', '".join(self.choices)))

    @property
    def desc(self):
        return self._desc

    def __repr__(self):
        return ("SwitchSpec(name='{}', default={}, choices={}, "
                "desc='{}')".format(self.name, self.default,
                                    self.choices, self.desc))



from builtins import object
import re
from copy import copy
from itertools import chain
from arcana.exceptions import (
    ArcanaUsageError, ArcanaInputError,
    ArcanaInputMissingMatchError, ArcanaNotBoundToStudyError)
from .base import BaseFileset, BaseField
from .item import Fileset, Field
from .collection import FilesetCollection, FieldCollection


class BaseInput(object):
    """
    Base class for Fileset and Field Input classes
    """

    def __init__(self, pattern, is_regex, order, from_study,
                 skip_missing=False, drop_if_missing=False,
                 fallback_to_default=False,
                 repository=None, study_=None, collection_=None):
        self._pattern = pattern
        self._is_regex = is_regex
        self._order = order
        self._from_study = from_study
        self._repository = repository
        self._skip_missing = skip_missing
        self._drop_if_missing = drop_if_missing
        self._fallback_to_default = fallback_to_default
        if skip_missing and fallback_to_default:
            raise ArcanaUsageError(
                "Cannot provide both mutually exclusive 'skip_missing' and "
                "'fallback_to_default' flags to {}".format(self))
        # Set when fallback_to_default is True and there are missing matches
        self._derivable = False
        self._fallback = None
        # study_ and collection_ are not intended to be provided to __init__
        # except when recreating when using initkwargs
        self._study = study_
        self._collection = collection_

    def __eq__(self, other):
        return (self.from_study == other.from_study and
                self.pattern == other.pattern and
                self.is_regex == other.is_regex and
                self.order == other.order and
                self._repository == other._repository and
                self._skip_missing == other._skip_missing and
                self._drop_if_missing == other._drop_if_missing and
                self._fallback_to_default == other._fallback_to_default)

    def __hash__(self):
        return (hash(self.from_study) ^
                hash(self.pattern) ^
                hash(self.is_regex) ^
                hash(self.order) ^
                hash(self._repository) ^
                hash(self._skip_missing) ^
                hash(self._drop_if_missing) ^
                hash(self._fallback_to_default))

    def initkwargs(self):
        dct = {}
        dct['from_study'] = self.from_study
        dct['pattern'] = self.pattern
        dct['order'] = self.order
        dct['is_regex'] = self.is_regex
        dct['study_'] = self._study
        dct['collection_'] = self._collection
        dct['skip_missing'] = self._skip_missing
        dct['drop_if_missing'] = self._drop_if_missing
        dct['fallback_to_default'] = self._fallback_to_default
        return dct

    @property
    def pattern(self):
        return self._pattern

    @property
    def spec_name(self):
        return self.name

    @property
    def derived(self):
        return self._from_study is not None

    @property
    def derivable(self):
        return self._derivable

    @property
    def from_study(self):
        return self._from_study

    @property
    def skip_missing(self):
        return self._skip_missing

    @property
    def drop_if_missing(self):
        return self._drop_if_missing

    @property
    def fallback_to_default(self):
        return self._fallback_to_default

    @property
    def study(self):
        if self._study is None:
            raise ArcanaNotBoundToStudyError(
                "{} is not bound to a study".format(self))
        return self._study

    @property
    def repository(self):
        if self._repository is None:
            if self._study is None:
                raise ArcanaUsageError(
                    "Cannot access repository of {} as it wasn't explicitly "
                    "provided and Input hasn't been bound to a study"
                    .format(self))
            repo = self._study.repository
        else:
            repo = self._repository
        return repo

    @property
    def collection(self):
        if self._collection is None:
            raise ArcanaNotBoundToStudyError(
                "{} has not been bound to a study".format(self))
        return self._collection

    @property
    def pipeline_getter(self):
        "For duck-typing with *Spec types"
        if not self.derivable:
            raise ArcanaUsageError(
                "There is no pipeline getter for {} because it doesn't "
                "fallback to a derived spec".format(self))
        return self._fallback.pipeline_getter

    @property
    def is_regex(self):
        return self._is_regex

    @property
    def order(self):
        return self._order

    def bind(self, study, spec_name=None, **kwargs):
        if spec_name is None:
            spec_name = self.spec_name
        if self._study == study:
            bound = self
        else:
            # Create copy and set study
            bound = copy(self)
            bound._study = study
            spec = study.data_spec(spec_name)
            # Use the default study repository if not explicitly
            # provided to match
            if self.fallback_to_default:
                if spec.derived:
                    bound._derivable = True
                elif spec.default is not None:
                    raise ArcanaUsageError(
                        "Cannot fallback to default for '{}' spec in {} as it "
                        " is not derived and doesn't have a default"
                        .format(self.name, study))
                # We don't want to add the bound copy to the study as that will
                # override the selector so we bind the fallback to the study
                # from the spec explicitly
                bound._fallback = spec.bind(study)
            # Match against tree
            if self._repository is None:
                repository = study.repository
            else:
                repository = self._repository
            with repository:
                tree = repository.cached_tree(
                    subject_ids=study.subject_ids,
                    visit_ids=study.visit_ids)
                bound._collection = bound.match(tree, spec=spec, **kwargs)
        return bound

    @property
    def prefixed_name(self):
        return self.name

    def nodes(self, tree):
        # Run the match against the tree
        if self.frequency == 'per_session':
            nodes = chain(*(s.sessions for s in tree.subjects))
        elif self.frequency == 'per_subject':
            nodes = tree.subjects
        elif self.frequency == 'per_visit':
            nodes = tree.visits
        elif self.frequency == 'per_study':
            nodes = [tree]
        else:
            assert False, "Unrecognised frequency '{}'".format(self.frequency)
        return nodes

    def _match(self, tree, item_cls, **kwargs):
        matches = []
        for node in self.nodes(tree):
            try:
                matches.append(self.match_node(node, **kwargs))
            except ArcanaInputMissingMatchError as e:
                if self._fallback is not None:
                    matches.append(self._fallback.collection.item(
                        subject_id=node.subject_id, visit_id=node.visit_id))
                elif self.skip_missing:
                    # Insert a non-existant item placeholder in-place of the
                    # the missing item
                    matches.append(item_cls(
                        self.name,
                        frequency=self.frequency,
                        subject_id=node.subject_id,
                        visit_id=node.visit_id,
                        repository=self.study.repository,
                        from_study=self.from_study,
                        exists=False,
                        **self._specific_kwargs))
                else:
                    raise e
        return matches

    def match_node(self, node, **kwargs):
        # Get names matching pattern
        matches = self._filtered_matches(node, **kwargs)
        # Filter matches by study name
        study_matches = [d for d in matches
                         if d.from_study == self.from_study]
        # Select the fileset from the matches
        if not study_matches:
            raise ArcanaInputMissingMatchError(
                "No matches found for {} in {} for study {}, however, found {}"
                .format(self, node, self.from_study,
                        ', '.join(str(m) for m in matches)))
        elif self.order is not None:
            try:
                match = study_matches[self.order]
            except IndexError:
                raise ArcanaInputMissingMatchError(
                    "Did not find {} named data matching pattern {}"
                    " (found {}) in {}".format(self.order, self.pattern,
                                               len(matches), node))
        elif len(study_matches) == 1:
            match = study_matches[0]
        else:
            raise ArcanaInputError(
                "Found multiple matches for {} in {} ({})"
                .format(self, node, ', '.join(str(m) for m in study_matches)))
        return match


class InputFileset(BaseInput, BaseFileset):
    """
    A pattern that describes a single fileset (typically acquired
    rather than generated but not necessarily) within each session.

    Parameters
    ----------
    spec_name : str
        The name of the fileset spec to match against
    format : FileFormat
        The file format used to store the fileset.
    pattern : str
        A regex pattern to match the fileset names with. Must match
        one and only one fileset per <frequency>. If None, the name
        is used instead.
    is_regex : bool
        Flags whether the pattern is a regular expression or not
    frequency : str
        One of 'per_session', 'per_subject', 'per_visit' and 'per_study',
        specifying whether the fileset is present for each session, subject,
        visit or project.
    id : int | None
        To be used to distinguish multiple filesets that match the
        pattern in the same session. The ID of the fileset within the
        session.
    order : int | None
        To be used to distinguish multiple filesets that match the
        pattern in the same session. The order of the fileset within the
        session. Based on the scan ID but is more robust to small
        changes to the IDs within the session if for example there are
        two scans of the same type taken before and after a task.
    dicom_tags : dct(str | str)
        To be used to distinguish multiple filesets that match the
        pattern in the same session. The provided DICOM values dicom
        header values must match exactly.
    from_study : str
        The name of the study that generated the derived fileset to match.
        Is used to determine the location of the filesets in the
        repository as the derived filesets and fields are grouped by
        the name of the study that generated them.
    skip_missing : bool
        If there is no fileset matching the selector for a node then pipelines
        that use it as an input, including downstream pipelines, will not be
        run for that node
    drop_if_missing : bool
        If there are missing filesets then drop the selector from the study
        input. Useful in the case where you want to provide selectors for the
        a list of inputs which may or may not be acquired for a range of
        studies
    fallback_to_default : bool
        If there is no fileset matching the selection for a node
        and corresponding data spec has a default or is a derived spec
        then fallback to the default or generate the derivative.
    repository : Repository | None
        The repository to draw the matches from, if not the main repository
        that is used to store the products of the current study.
    """

    is_spec = False

    def __init__(self, spec_name, pattern=None, format=None, # @ReservedAssignment @IgnorePep8
                 frequency='per_session', id=None,  # @ReservedAssignment
                 order=None, dicom_tags=None, is_regex=False, from_study=None,
                 skip_missing=False, drop_if_missing=False,
                 fallback_to_default=False, repository=None, study_=None,
                 collection_=None):
        BaseFileset.__init__(self, spec_name, format, frequency)
        BaseInput.__init__(self, pattern, is_regex, order,
                              from_study, skip_missing, drop_if_missing,
                              fallback_to_default, repository, study_,
                              collection_)
        if dicom_tags is not None and format.name != 'dicom':
            raise ArcanaUsageError(
                "Cannot use 'dicom_tags' kwarg with non-DICOM "
                "format ({})".format(format))
        self._dicom_tags = dicom_tags
        if order is not None and id is not None:
            raise ArcanaUsageError(
                "Cannot provide both 'order' and 'id' to a fileset"
                "match")
        self._id = str(id) if id is not None else id
        self._check_args()

    def __eq__(self, other):
        return (BaseFileset.__eq__(self, other) and
                BaseInput.__eq__(self, other) and
                self.dicom_tags == other.dicom_tags and
                self.id == other.id)

    def __hash__(self):
        return (BaseFileset.__hash__(self) ^
                BaseInput.__hash__(self) ^
                hash(self.dicom_tags) ^
                hash(self.id))

    def _check_args(self):
        if self._pattern is None and self._id is None:
            raise ArcanaUsageError(
                "Either 'pattern' or 'id' need to be provided to "
                "InputFileset constructor")

    def initkwargs(self):
        dct = BaseFileset.initkwargs(self)
        dct.update(BaseInput.initkwargs(self))
        dct['dicom_tags'] = self.dicom_tags
        dct['id'] = self.id
        return dct

    def __repr__(self):
        return ("{}(name='{}', format={}, frequency={}, pattern={}, "
                "is_regex={}, order={}, id={}, dicom_tags={}, "
                "from_study={})"
                .format(self.__class__.__name__, self.name, self._format,
                        self.frequency, self._pattern, self.is_regex,
                        self.order, self.id, self.dicom_tags,
                        self._from_study))

    def match(self, tree, spec, **kwargs):
        if self.format is not None:
            candidate_formats = [self.format]
        else:
            candidate_formats = spec.valid_formats
        # Run the match against the tree
        return FilesetCollection(self.name,
                                 self._match(tree, Fileset, **kwargs),
                                 frequency=self.frequency,
                                 candidate_formats=candidate_formats)

    @property
    def id(self):
        return self._id

    @property
    def format(self):
        if self._format is None:
            try:
                format = self.collection.format  # @ReservedAssignment
            except ArcanaNotBoundToStudyError:
                format = None  # @ReservedAssignment
        else:
            format = self._format  # @ReservedAssignment
        return format

    @property
    def dicom_tags(self):
        return self._dicom_tags

    def _filtered_matches(self, node):
        if self.pattern is not None:
            if self.is_regex:
                pattern_re = re.compile(self.pattern)
                matches = [f for f in node.filesets
                           if pattern_re.match(f.basename)]
            else:
                matches = [f for f in node.filesets
                           if f.basename == self.pattern]
        else:
            matches = list(node.filesets)
        if not matches:
            raise ArcanaInputMissingMatchError(
                "Did not find any matches for {} in {}, found:\n{}"
                .format(self, node,
                        '\n'.join(str(f) for f in node.filesets)))
        if self.id is not None:
            filtered = [d for d in matches if d.id == self.id]
            if not filtered:
                raise ArcanaInputMissingMatchError(
                    "Did not find filesets names matching pattern {} "
                    "with an id of {} (found {}) in {}".format(
                        self.pattern, self.id,
                        ', '.join(str(m) for m in matches), node))
            matches = filtered
        if self.format is not None:
            format_matches = [f for f in matches if self.format.matches(f)]
            if not format_matches:
                for f in matches:
                    self.format.matches(f)
                raise ArcanaInputMissingMatchError(
                    "Did not find any filesets that match the file format "
                    "specified by {} in {}, found:\n{}"
                    .format(self, node, '\n'.join(str(f) for f in matches)))
            matches = format_matches
        # Filter matches by dicom tags
        if self.dicom_tags is not None:
            filtered = []
            for fileset in matches:
                keys, ref_values = zip(*self.dicom_tags.items())
                values = tuple(self.format.dicom_values(fileset, keys))
                if ref_values == values:
                    filtered.append(fileset)
            if not filtered:
                raise ArcanaInputMissingMatchError(
                    "Did not find filesets names matching pattern {}"
                    "that matched DICOM tags {} (found {}) in {}"
                    .format(self.pattern, self.dicom_tags,
                            ', '.join(str(m) for m in matches), node))
            matches = filtered
        return matches

    def cache(self):
        """
        Forces the cache of the input fileset. Can be useful for before running
        a workflow that will use many concurrent jobs/processes to source data
        from remote repository, to force the download to be done linearly and
        avoid DOSing the host
        """
        for item in self.collection:
            item.get()

    @property
    def _specific_kwargs(self):
        return {'format': self.format}


class InputField(BaseInput, BaseField):
    """
    A pattern that matches a single field (typically acquired rather than
    generated but not necessarily) in each session.

    Parameters
    ----------
    spec_name : str
        The name of the field spec to match against.
    pattern : str
        A regex pattern to match the field names with. Must match
        one and only one fileset per <frequency>. If None, the name
        is used instead.
    dtype : type | None
        The datatype of the value. Can be one of (float, int, str). If None
        then the dtype is taken from the FieldSpec that it is bound to
    is_regex : bool
        Flags whether the pattern is a regular expression or not
    frequency : str
        One of 'per_session', 'per_subject', 'per_visit' and 'per_study',
        specifying whether the fileset is present for each session, subject,
        visit or project.
    order : int | None
        To be used to distinguish multiple filesets that match the
        pattern in the same session. The order of the fileset within the
        session. Based on the scan ID but is more robust to small
        changes to the IDs within the session if for example there are
        two scans of the same type taken before and after a task.
    from_study : str
        The name of the study that generated the derived field to match.
        Is used to determine the location of the fields in the
        repository as the derived filesets and fields are grouped by
        the name of the study that generated them.
    skip_missing : bool
        If there is no field matching the selector for a node then pipelines
        that use it as an input, including downstream pipelines, will not be
        run for that node
    drop_if_missing : bool
        If there are missing filesets then drop the selector from the study
        input. Useful in the case where you want to provide selectors for the
        a list of inputs which may or may not be acquired for a range of
        studies
    fallback_to_default : bool
        If the there is no fileset/field matching the selection for a node
        and corresponding data spec has a default or is a derived spec,
        then fallback to the default or generate the derivative.
    repository : Repository | None
        The repository to draw the matches from, if not the main repository
        that is used to store the products of the current study.
    """

    is_spec = False

    def __init__(self, spec_name, pattern, dtype=None, frequency='per_session',
                 order=None, is_regex=False, from_study=None,
                 skip_missing=False, drop_if_missing=False,
                 fallback_to_default=False, repository=None, study_=None,
                 collection_=None):
        BaseField.__init__(self, spec_name, dtype, frequency)
        BaseInput.__init__(self, pattern, is_regex, order,
                              from_study, skip_missing, drop_if_missing,
                              fallback_to_default, repository, study_,
                              collection_)

    def __eq__(self, other):
        return (BaseField.__eq__(self, other) and
                BaseInput.__eq__(self, other))

    def match(self, tree, spec=None, **kwargs):  # @UnusedVariable
        # Run the match against the tree
        return FieldCollection(self.name,
                               self._match(tree, Field, **kwargs),
                               frequency=self.frequency,
                               dtype=self.dtype)

    @property
    def dtype(self):
        if self._dtype is None:
            try:
                dtype = self.study.data_spec(self.name).dtype
            except ArcanaNotBoundToStudyError:
                dtype = None
        else:
            dtype = self._dtype
        return dtype

    def __hash__(self):
        return (BaseField.__hash__(self) ^ BaseInput.__hash__(self))

    def initkwargs(self):
        dct = BaseField.initkwargs(self)
        dct.update(BaseInput.initkwargs(self))
        return dct

    def _filtered_matches(self, node):
        if self.is_regex:
            pattern_re = re.compile(self.pattern)
            matches = [f for f in node.fields
                       if pattern_re.match(f.name)]
        else:
            matches = [f for f in node.fields
                       if f.name == self.pattern]
        if self.from_study is not None:
            matches = [f for f in matches
                       if f.from_study == self.from_study]
        if not matches:
            raise ArcanaInputMissingMatchError(
                "Did not find any matches for {} in {}, found:\n{}"
                .format(self, node, '\n'.join(f.name for f in node.fields)))
        return matches

    def __repr__(self):
        return ("{}(name='{}', dtype={}, frequency={}, pattern={}, "
                "is_regex={}, order={}, from_study={})"
                .format(self.__class__.__name__, self.name, self._dtype,
                        self.frequency, self._pattern, self.is_regex,
                        self.order, self._from_study))

    @property
    def _specific_kwargs(self):
        return {'dtype': self.dtype}
